@@ -17,30 +17,40 @@
  */
 
 #include "xchat.h"
-#undef USE_GNOME
 #include "gtkutil.h"
 #include <fcntl.h>
 #include <time.h>
 
 
+extern unsigned char *strip_color (unsigned char *text);
 extern int tcp_send_len (struct server *serv, char *buf, int len);
 extern int tcp_send (struct server *serv, char *buf);
 extern void path_part (char *file, char *path);
+extern int match (char *mask, char *str);
 
 
 void
 chanlist_addentry (struct server *serv, char *chan, char *users, char *topic)
 {
-   gchar *new[1][3];
+   gchar *nnew[3];
+   char *wild;
 
    if (atoi (users) < serv->chanlist_minusers)
       return;
 
-   new[0][0] = chan;
-   new[0][1] = users;
-   new[0][2] = topic;
+   wild = gtk_entry_get_text (GTK_ENTRY (serv->chanlist_wild));
+   if (wild && wild[0])
+   {
+      if (!match (wild, chan))
+         return;
+   }
+   nnew[0] = chan;
+   nnew[1] = users;
+   nnew[2] = strip_color (topic);
 
-   gtk_clist_append (GTK_CLIST (serv->chanlist_list), new[0]);
+   gtk_clist_append (GTK_CLIST (serv->chanlist_list), nnew);
+
+   free (nnew[2]);
 }
 
 void
@@ -48,17 +58,18 @@ chanlist_join (GtkWidget * wid, struct server *serv)
 {
    int row;
    char *chan;
+   char tbuf[256];
 
    row = gtkutil_clist_selection (serv->chanlist_list);
    if (row != -1)
    {
       gtk_clist_get_text (GTK_CLIST (serv->chanlist_list), row, 0, &chan);
-      if (serv->connected)
+      if (serv->connected && (strcmp (chan, "*") != 0))
       {
-         char tbuf[128];
-         sprintf (tbuf, "JOIN %s\r\n", chan);
+         snprintf (tbuf, sizeof tbuf, "JOIN %s\r\n", chan);
          tcp_send (serv, tbuf);
-      }
+      } else
+         gdk_beep ();
    }
 }
 
@@ -114,9 +125,13 @@ chanlist_save (GtkWidget * wid, struct server *serv)
 void
 chanlist_refresh (GtkWidget * wid, struct server *serv)
 {
-   gtk_clist_clear (GTK_CLIST (serv->chanlist_list));
    if (serv->connected)
+   {
+      gtk_clist_clear (GTK_CLIST (serv->chanlist_list));
+      gtk_widget_set_sensitive (serv->chanlist_refresh, FALSE);
       tcp_send_len (serv, "LIST\r\n", 6);
+   } else
+      gtkutil_simpledialog ("Not connected.");
 }
 
 void
@@ -152,7 +167,7 @@ chanlist_opengui (struct server *serv)
 
    if (serv->chanlist_window)
    {
-      gtk_widget_destroy (serv->chanlist_window);
+      gdk_window_show (serv->chanlist_window->window);
       return;
    }
    snprintf (tbuf, sizeof tbuf, "X-Chat: Channel List (%s)", serv->servername);
@@ -174,21 +189,35 @@ chanlist_opengui (struct server *serv)
    gtk_box_pack_start (GTK_BOX (hbox), wid, 0, 0, 0);
    gtk_widget_show (wid);
 
+   if (!serv->chanlist_minusers)
+      serv->chanlist_minusers = 3;
    wid = gtk_entry_new_with_max_length (6);
    gtk_widget_set_usize (wid, 40, 0);
-   gtk_entry_set_text (GTK_ENTRY (wid), "3");
+   sprintf (tbuf, "%d", serv->chanlist_minusers);
+   gtk_entry_set_text (GTK_ENTRY (wid), tbuf);
    gtk_signal_connect (GTK_OBJECT (wid), "changed",
                        GTK_SIGNAL_FUNC (chanlist_minusers), serv);
    gtk_box_pack_start (GTK_BOX (hbox), wid, 0, 0, 0);
    gtk_widget_show (wid);
-   serv->chanlist_minusers = 3;
+
+   wid = gtk_label_new ("Wildcard Match:");
+   gtk_widget_set_usize (wid, 125, 0);
+   gtk_box_pack_start (GTK_BOX (hbox), wid, 0, 0, 0);
+   gtk_widget_show (wid);
+
+   wid = gtk_entry_new_with_max_length (32);
+   gtk_widget_set_usize (wid, 155, 0);
+   gtk_entry_set_text (GTK_ENTRY (wid), "*");
+   gtk_box_pack_start (GTK_BOX (hbox), wid, 0, 0, 0);
+   gtk_widget_show (wid);
+   serv->chanlist_wild = wid;
 
    serv->chanlist_list = gtkutil_clist_new (3, titles, vbox, GTK_POLICY_ALWAYS,
                 chanlist_row_selected,
                       (gpointer) serv,
           0, 0, GTK_SELECTION_BROWSE);
    gtk_clist_set_column_width (GTK_CLIST (serv->chanlist_list), 0, 90);
-   gtk_clist_set_column_width (GTK_CLIST (serv->chanlist_list), 1, 50);
+   gtk_clist_set_column_width (GTK_CLIST (serv->chanlist_list), 1, 45);
    gtk_clist_set_column_width (GTK_CLIST (serv->chanlist_list), 2, 165);
    gtk_clist_set_auto_sort (GTK_CLIST (serv->chanlist_list), 1);
 
@@ -196,9 +225,13 @@ chanlist_opengui (struct server *serv)
    gtk_box_pack_end (GTK_BOX (vbox), hbox, 0, 0, 0);
    gtk_widget_show (hbox);
 
-   gtkutil_button_new ("Refresh the list", 0, 0, chanlist_refresh, serv, hbox);
-   gtkutil_button_new ("Save the list", 0, 0, chanlist_save, serv, hbox);
-   gtkutil_button_new ("Join Channel", 0, 0, chanlist_join, serv, hbox);
+   serv->chanlist_refresh =
+      gtkutil_button (serv->chanlist_window, GNOME_STOCK_PIXMAP_REFRESH,
+                      "Refresh the list", chanlist_refresh, serv, hbox);
+   gtkutil_button (serv->chanlist_window, GNOME_STOCK_PIXMAP_SAVE,
+                   "Save the list", chanlist_save, serv, hbox);
+   gtkutil_button (serv->chanlist_window, GNOME_STOCK_PIXMAP_JUMP_TO,
+                   "Join Channel", chanlist_join, serv, hbox);
 
    gtk_widget_show (serv->chanlist_window);
 }
