@@ -18,7 +18,6 @@
  * Wayne Conrad, 3 Apr 1999: Color-coded DCC file transfer status windows
  */
 
-#include "style.h"
 #include "xchat.h"
 #include "util.h"
 #include <sys/stat.h>
@@ -28,7 +27,7 @@
 #include <time.h>
 #include <errno.h>
 #include "dcc.h"
-#include "draw.h"
+/*#include "draw.h"*/
 #include "gtkutil.h"
 #include "plugin.h"
 #include <time.h>
@@ -38,8 +37,7 @@
 
 extern GdkColor colors[];
 
-static char *dcctypes[] =
-{"SEND", "RECV", "CHAT", "CHAT", "DRAW", "DRAW"};
+static char *dcctypes[] = {"SEND", "RECV", "CHAT", "CHAT",};/* "DRAW", "DRAW"};*/
 
 
 /* The name and color of each dcc status. */
@@ -109,6 +107,11 @@ extern void url_checkurl (char *buf);
 extern void channel_action (struct session *sess, char *tbuf, char *chan, char *from, char *text, char fromme);
 extern struct session *find_session_from_channel (char *chan, struct server *serv);
 
+#ifdef USE_PERL
+/* perl.c */
+
+int perl_dcc_chat (struct session *sess, struct server *serv, char *buf);
+#endif
 
 
 int
@@ -231,9 +234,12 @@ dcc_close (struct DCC *dcc, int stat, int destroy)
    char type = dcc->type;
    if (dcc->sok != -1)
    {
+      if (dcc->wiotag != -1)
+         gdk_input_remove (dcc->wiotag);
       gdk_input_remove (dcc->iotag);
       close (dcc->sok);
       dcc->sok = -1;
+      dcc->wiotag = -1;
    }
    if (dcc->fp != -1)
    {
@@ -241,11 +247,11 @@ dcc_close (struct DCC *dcc, int stat, int destroy)
       dcc->fp = -1;
    }
    dcc->stat = stat;
-   if (type == TYPE_DRAWSEND || type == TYPE_DRAWRECV)
+/*   if (type == TYPE_DRAWSEND || type == TYPE_DRAWRECV)
    {
       if (dcc->dccdraw)
          dcc_close_draw_window (dcc);
-   }
+   }*/
    if (dcc->dccchat)
    {
       free (dcc->dccchat);
@@ -310,7 +316,7 @@ dcc_write_chat (char *nick, char *text)
 static void
 dcc_read_chat (struct DCC *dcc, gint sok)
 {
-   int i;
+   int i, skip;
    long len;
    char tbuf[1226];
    char lbuf[1026];
@@ -325,8 +331,12 @@ dcc_read_chat (struct DCC *dcc, gint sok)
             if (errno == EAGAIN || errno == EWOULDBLOCK)
                return;
          }
-         EMIT_SIGNAL (XP_TE_DCCCHATF, dcc->serv->front_session, dcc->nick, NULL,
-                      NULL, NULL, 0);
+         {
+           char dccport[4];
+           sprintf(dccport,"%04ld",dcc->port);
+           EMIT_SIGNAL (XP_TE_DCCCHATF, dcc->serv->front_session, dcc->nick, 
+                 inet_ntoa (dcc->SAddr.sin_addr), dccport, NULL, 0);
+         }
          dcc_close (dcc, STAT_FAILED, FALSE);
          return;
       }
@@ -340,18 +350,42 @@ dcc_read_chat (struct DCC *dcc, gint sok)
             break;
          case '\n':
             dcc->dccchat->linebuf[dcc->dccchat->pos] = 0;
+#ifdef USE_PERL
+            {
+	      char *host_n_nick_n_message;
+
+	      host_n_nick_n_message=malloc(
+	        strlen(dcc->nick)+strlen(dcc->dccchat->linebuf)+26
+	      );
+              sprintf (
+	        host_n_nick_n_message,
+		"%s %ld %s: %s",
+		inet_ntoa (dcc->SAddr.sin_addr),
+		dcc->port,
+		dcc->nick,
+		dcc->dccchat->linebuf
+	      );
+            skip = perl_dcc_chat(find_session_from_channel (dcc->nick, dcc->serv), dcc->serv, host_n_nick_n_message);
+	      free(host_n_nick_n_message);
+	    }
+#else
+		skip = 0;
+#endif
             url_checkurl (dcc->dccchat->linebuf);
 
-            if (dcc->dccchat->linebuf[0] == 1 && !strncasecmp (dcc->dccchat->linebuf + 1, "ACTION", 6))
-            {
-               char *po = strchr (dcc->dccchat->linebuf + 8, '\001');
-               if (po)
-                  po[0] = 0;
-               channel_action (find_session_from_channel (dcc->nick, dcc->serv), tbuf, dcc->serv->nick, dcc->nick, dcc->dccchat->linebuf + 8, FALSE);
-            } else
-            {
-               private_msg (dcc->serv, tbuf, dcc->nick, "", dcc->dccchat->linebuf);
-            }
+		if (!skip)
+		{
+         	   if (dcc->dccchat->linebuf[0] == 1 && !strncasecmp (dcc->dccchat->linebuf + 1, "ACTION", 6))
+         	   {
+            	char *po = strchr (dcc->dccchat->linebuf + 8, '\001');
+            	if (po)
+               	   po[0] = 0;
+            	channel_action (find_session_from_channel (dcc->nick, dcc->serv), tbuf, dcc->serv->nick, dcc->nick, dcc->dccchat->linebuf + 8, FALSE);
+         	   } else
+         	   {
+            	private_msg (dcc->serv, tbuf, dcc->nick, "", dcc->dccchat->linebuf);
+         	   }
+		}
             dcc->pos += dcc->dccchat->pos;
             dcc->dccchat->pos = 0;
             update_dcc_chat_window ();
@@ -378,11 +412,7 @@ dcc_read (struct DCC *dcc, gint sok)
    {
       if (dcc->resumable)
       {
-#ifdef __EMX__                  /* OS/2 needs O_BINARY */
-         dcc->fp = open (dcc->destfile, O_BINARY | O_WRONLY | O_APPEND);
-#else
-         dcc->fp = open (dcc->destfile, O_WRONLY | O_APPEND);
-#endif
+         dcc->fp = open (dcc->destfile, O_WRONLY | O_APPEND | OFLAGS);
          dcc->pos = dcc->resumable;
          dcc->ack = dcc->resumable;
       } else
@@ -407,11 +437,7 @@ dcc_read (struct DCC *dcc, gint sok)
             EMIT_SIGNAL (XP_TE_DCCRENAME, dcc->serv->front_session,
                          dcc->destfile, buf, NULL, NULL, 0);
          }
-#ifdef __EMX__
-         dcc->fp = open (dcc->destfile, O_BINARY | O_TRUNC | O_WRONLY | O_CREAT, prefs.dccpermissions);
-#else
-         dcc->fp = open (dcc->destfile, O_TRUNC | O_WRONLY | O_CREAT, prefs.dccpermissions);
-#endif
+         dcc->fp = open (dcc->destfile, OFLAGS | O_TRUNC | O_WRONLY | O_CREAT, prefs.dccpermissions);
       }
    }
    if (dcc->fp == -1)
@@ -443,14 +469,15 @@ dcc_read (struct DCC *dcc, gint sok)
 
          write (dcc->fp, buf, n);
          dcc->pos += n;
-         pos = htonl (dcc->pos);
+         pos = htonl (dcc->pos - dcc->resumable);
          send (dcc->sok, (char *) &pos, 4, 0);
 
          dcc->lasttime = time (0);
          sec = dcc->lasttime - dcc->starttime;
          if (sec < 1)
             sec = 1;
-         dcc->cps = dcc->pos / sec;
+         dcc->cps = (dcc->pos - dcc->resumable) / sec;
+         dcc->perc = (gfloat) ((dcc->pos * 100.00) / dcc->size);
 
          if (dcc->pos >= dcc->size)
          {
@@ -509,19 +536,19 @@ dcc_connect_finished (struct DCC *dcc, gint sok)
       dcc->dccchat->pos = 0;
       break;
 
-   case TYPE_DRAWRECV:
+/*   case TYPE_DRAWRECV:
       dcc->iotag = gdk_input_add (dcc->sok, GDK_INPUT_READ | GDK_INPUT_EXCEPTION,
                                   (GdkInputFunction) dcc_read_draw, dcc);
       dcc_open_draw_window (dcc);
-      break;
+      break;*/
    }
    update_dcc_window (dcc->type);
    dcc->starttime = time (0);
    dcc->lasttime = dcc->starttime;
 
-   sprintf (host, "%s:%ld", inet_ntoa (dcc->SAddr.sin_addr), dcc->port);
-   EMIT_SIGNAL (XP_TE_DCCCON, dcc->serv->front_session, dcctypes[(int) dcc->type]
-           ,dcc->nick, host, "to", 0);
+   snprintf (host, sizeof host, "%s:%ld", inet_ntoa (dcc->SAddr.sin_addr), dcc->port);
+   EMIT_SIGNAL (XP_TE_DCCCON, dcc->serv->front_session, dcctypes[(int) dcc->type],
+			dcc->nick, host, "to", 0);
 }
 
 void
@@ -545,33 +572,29 @@ dcc_connect (struct session *sess, struct DCC *dcc)
 static void
 update_dcc_recv (struct DCC *dcc)
 {
-   char pos[14], cps[14];
+   char pos[14], cps[14], perc[14];
    gint row;
-#ifdef USE_GNOME
-   char file[512];
-#endif
 
    if (!dccrwin.window)
       return;
 
    row = gtk_clist_find_row_from_data (GTK_CLIST (dccrwin.list), (gpointer) dcc);
 
-   sprintf (pos, "%lu", dcc->pos);
-   sprintf (cps, "%lu", dcc->cps);
+   snprintf (pos, sizeof (pos), "%lu", dcc->pos);
+   snprintf (cps, sizeof (cps), "%lu", dcc->cps);
+   snprintf (perc, sizeof (perc), "%.0f%%", dcc->perc);
 
    gtk_clist_freeze (GTK_CLIST (dccrwin.list));
-   gtk_clist_set_text
-      (GTK_CLIST (dccrwin.list), row, 0, dccstat[(int) dcc->stat].name);
+   gtk_clist_set_text (GTK_CLIST (dccrwin.list), row, 0, dccstat[(int) dcc->stat].name);
    gtk_clist_set_text (GTK_CLIST (dccrwin.list), row, 3, pos);
-   gtk_clist_set_text (GTK_CLIST (dccrwin.list), row, 4, cps);
-   gtk_clist_set_foreground
-      (GTK_CLIST (dccrwin.list), row, colors + dccstat[(int) dcc->stat].color);
+   gtk_clist_set_text (GTK_CLIST (dccrwin.list), row, 4, perc);
+   gtk_clist_set_text (GTK_CLIST (dccrwin.list), row, 5, cps);
+   if (dccstat[(int) dcc->stat].color != 1)
+	gtk_clist_set_foreground
+   	   (GTK_CLIST (dccrwin.list), row, colors + dccstat[(int) dcc->stat].color);
 #ifdef USE_GNOME
    if (dcc->stat == STAT_DONE)
-   {
-      snprintf (file, sizeof (file), "%s/%s", prefs.dccdir, dcc->file);
-      gtk_clist_set_text (GTK_CLIST (dccrwin.list), row, 6, (char *) gnome_mime_type_of_file (file));
-   }
+      gtk_clist_set_text (GTK_CLIST (dccrwin.list), row, 7, (char *) gnome_mime_type_of_file (dcc->destfile));
 #endif
    gtk_clist_thaw (GTK_CLIST (dccrwin.list));
 }
@@ -579,7 +602,7 @@ update_dcc_recv (struct DCC *dcc)
 static void
 update_dcc_send (struct DCC *dcc)
 {
-   char pos[14], cps[14], ack[14];
+   char pos[14], cps[14], ack[14], perc[14];
    gint row;
 
    if (!dccswin.window)
@@ -587,23 +610,76 @@ update_dcc_send (struct DCC *dcc)
 
    row = gtk_clist_find_row_from_data (GTK_CLIST (dccswin.list), (gpointer) dcc);
 
-   sprintf (pos, "%lu", dcc->pos);
-   sprintf (cps, "%lu", dcc->cps);
-   sprintf (ack, "%lu", dcc->ack);
+   snprintf (pos, sizeof (pos), "%lu", dcc->pos);
+   snprintf (cps, sizeof (cps), "%lu", dcc->cps);
+   snprintf (ack, sizeof (ack), "%lu", dcc->ack + dcc->resumable);
+   snprintf (perc, sizeof (perc), "%.0f%%", dcc->perc);
 
    gtk_clist_freeze (GTK_CLIST (dccswin.list));
-   gtk_clist_set_text
-      (GTK_CLIST (dccswin.list), row, 0, dccstat[(int) dcc->stat].name);
-   gtk_clist_set_foreground
-      (GTK_CLIST (dccswin.list), row, colors + dccstat[(int) dcc->stat].color);
+   gtk_clist_set_text (GTK_CLIST (dccswin.list), row, 0, dccstat[(int) dcc->stat].name);
+   if (dccstat[(int) dcc->stat].color != 1)
+	gtk_clist_set_foreground
+   	   (GTK_CLIST (dccswin.list), row, colors + dccstat[(int) dcc->stat].color);
    gtk_clist_set_text (GTK_CLIST (dccswin.list), row, 3, pos);
    gtk_clist_set_text (GTK_CLIST (dccswin.list), row, 4, ack);
-   gtk_clist_set_text (GTK_CLIST (dccswin.list), row, 5, cps);
+   gtk_clist_set_text (GTK_CLIST (dccswin.list), row, 5, perc);
+   gtk_clist_set_text (GTK_CLIST (dccswin.list), row, 6, cps);
    gtk_clist_thaw (GTK_CLIST (dccswin.list));
 }
 
-static int
-dcc_read_ack (struct DCC *dcc, char *buf)
+static void
+dcc_send_data (struct DCC *dcc, gint sok)
+{
+   char buf[2048];
+   long len, sec, sent;
+
+   if (dcc->pos < dcc->size)
+   {
+      lseek (dcc->fp, dcc->pos, SEEK_SET);
+      len = read (dcc->fp, buf, sizeof buf);
+
+      sent = send (dcc->sok, buf, len, 0);
+
+      if (sent < 0 && errno != EWOULDBLOCK)
+      {
+         EMIT_SIGNAL (XP_TE_DCCSENDFAIL, dcc->serv->front_session,
+                      file_part (dcc->file), dcc->nick, NULL, NULL, 0);
+         dcc_close (dcc, STAT_FAILED, FALSE);
+         return;
+      }
+      dcc->lasttime = time (0);
+      sec = dcc->lasttime - dcc->starttime;
+      if (sec < 1)
+         sec = 1;
+      dcc->cps = (dcc->pos - dcc->resumable) / sec;
+      if (sent > 0)
+         dcc->pos += sent;
+      dcc->perc = (gfloat) ((dcc->pos * 100.00) / dcc->size);
+
+
+   }
+   if (dcc->pos >= dcc->size && (dcc->ack + dcc->resumable) >= dcc->size)
+   {
+      dcc->perc = 100.0;
+      dcc_close (dcc, STAT_DONE, FALSE);
+      snprintf (buf, sizeof (buf), "%ld", dcc->cps);
+
+      EMIT_SIGNAL (XP_TE_DCCSENDCOMP, dcc->serv->front_session,
+                   file_part (dcc->file), dcc->nick, buf, NULL, 0);
+   } else
+   {
+      if (dcc->pos >= (dcc->oldpos + DCC_UPDATE_WINDOW) ||
+          dcc->ack >= (dcc->oldack + DCC_UPDATE_WINDOW))
+      {
+         dcc->oldack = dcc->ack;
+         dcc->oldpos = dcc->pos;
+         update_dcc_send (dcc);
+      }
+   }
+}
+
+static void
+dcc_read_ack (struct DCC *dcc, gint sok)
 {
    int len;
    guint32 ack;
@@ -614,84 +690,23 @@ dcc_read_ack (struct DCC *dcc, char *buf)
       if (len < 0)
       {
          if (errno == EWOULDBLOCK)
-            return 1;
+            return;
       }
       EMIT_SIGNAL (XP_TE_DCCSENDFAIL, dcc->serv->front_session,
                    file_part (dcc->file), dcc->nick, NULL, NULL, 0);
       dcc_close (dcc, STAT_FAILED, FALSE);
-      return 1;
+      return;
    }
    if (len < 4)
-      return 1;
+      return;
    recv (dcc->sok, (char *) &ack, 4, 0);
    dcc->ack = ntohl (ack);
-   return 0;
-}
 
-static void
-dcc_send_data (struct DCC *dcc, gint sok)
-{
-   char buf[2048];
-   long len, sec, sent;
-
-   while (1)
+   if (!dcc->fastsend)
    {
-      if (sok != -1)
-      {
-         if (dcc_read_ack (dcc, buf))
-            return;
-      } else
-      {
-         sok = dcc->sok;
-      }
-
-      if (dcc->pos < dcc->size)
-      {
-
-         if (!prefs.fastdccsend)
-         {
-            if ((dcc->ack + 4096) < dcc->pos)
-               return;
-         }
-         lseek (dcc->fp, dcc->pos, SEEK_SET);
-         len = read (dcc->fp, buf, sizeof buf);
-
-         sent = send (dcc->sok, buf, len, 0);
-
-         if (sent < 0 && errno != EWOULDBLOCK)
-         {
-            EMIT_SIGNAL (XP_TE_DCCSENDFAIL, dcc->serv->front_session,
-                         file_part (dcc->file), dcc->nick, NULL, NULL, 0);
-            dcc_close (dcc, STAT_FAILED, FALSE);
-            return;
-         }
-         dcc->lasttime = time (0);
-         sec = dcc->lasttime - dcc->starttime;
-         if (sec < 1)
-            sec = 1;
-         dcc->cps = dcc->pos / sec;
-         if (sent > 0)
-            dcc->pos += sent;
-
-      }
-      if (dcc->pos >= dcc->size && dcc->ack >= dcc->size)
-      {
-         dcc_close (dcc, STAT_DONE, FALSE);
-         snprintf (buf, sizeof (buf), "%ld", dcc->cps);
-
-         EMIT_SIGNAL (XP_TE_DCCSENDCOMP, dcc->serv->front_session,
-                      file_part (dcc->file), dcc->nick, buf, NULL, 0);
+      if ((dcc->ack + dcc->resumable) < dcc->pos)
          return;
-      } else
-      {
-         if (dcc->pos >= (dcc->oldpos + DCC_UPDATE_WINDOW) ||
-             dcc->ack >= (dcc->oldack + DCC_UPDATE_WINDOW))
-         {
-            dcc->oldack = dcc->ack;
-            dcc->oldpos = dcc->pos;
-            update_dcc_send (dcc);
-         }
-      }
+      dcc_send_data (dcc, dcc->sok);
    }
 }
 
@@ -716,14 +731,19 @@ dcc_accept (struct DCC *dcc, gint sokk)
    dcc->addr = ntohl (CAddr.sin_addr.s_addr);
    memcpy ((char *) &dcc->SAddr.sin_addr, (char *) &CAddr.sin_addr, sizeof (struct in_addr));
    dcc->stat = STAT_ACTIVE;
-   dcc->starttime = time (0);
-   dcc->lasttime = dcc->starttime;
+   dcc->lasttime = dcc->starttime = time (0);
+   dcc->fastsend = prefs.fastdccsend;
    switch (dcc->type)
    {
    case TYPE_SEND:
+      if (dcc->fastsend)
+      {
+         dcc->wiotag = gdk_input_add (sok, GDK_INPUT_WRITE,
+                                      (GdkInputFunction) dcc_send_data, dcc);
+      }
       dcc->iotag = gdk_input_add (sok, GDK_INPUT_READ | GDK_INPUT_EXCEPTION,
-                                  (GdkInputFunction) dcc_send_data, dcc);
-      dcc_send_data (dcc, -1);
+                                  (GdkInputFunction) dcc_read_ack, dcc);
+      dcc_send_data (dcc, dcc->sok);
       break;
 
    case TYPE_CHATSEND:
@@ -733,16 +753,16 @@ dcc_accept (struct DCC *dcc, gint sokk)
       dcc->dccchat->pos = 0;
       break;
 
-   case TYPE_DRAWSEND:
+/*   case TYPE_DRAWSEND:
       dcc->iotag = gdk_input_add (dcc->sok, GDK_INPUT_READ | GDK_INPUT_EXCEPTION,
                                   (GdkInputFunction) dcc_read_draw, dcc);
       dcc_open_draw_window (dcc);
-      break;
+      break;*/
    }
 
    update_dcc_window (dcc->type);
 
-   sprintf (host, "%s:%ld", inet_ntoa (dcc->SAddr.sin_addr), dcc->port);
+   snprintf (host, sizeof host, "%s:%ld", inet_ntoa (dcc->SAddr.sin_addr), dcc->port);
    EMIT_SIGNAL (XP_TE_DCCCON, dcc->serv->front_session, dcctypes[(int) dcc->type],
           dcc->nick, host, "from", 0);
 }
@@ -765,6 +785,13 @@ dcc_listen_init (struct DCC *dcc)
    SAddr.sin_family = AF_INET;
    SAddr.sin_port = 0;
 
+   /* this is safe - we can't have bound to a local_ip if we're getting from
+      server */
+   if (prefs.local_ip != 0)
+   {
+      /* bind to internet facing side */
+      SAddr.sin_addr.s_addr = prefs.local_ip;
+   }
    bind (dcc->sok, (struct sockaddr *) &SAddr, sizeof (SAddr));
 
    len = sizeof (SAddr);
@@ -773,9 +800,15 @@ dcc_listen_init (struct DCC *dcc)
    memcpy ((char *) &dcc->SAddr, (char *) &SAddr, sizeof (struct sockaddr_in));
    dcc->port = ntohs (SAddr.sin_port);
    if (prefs.my_ip != 0)
+   {
       dcc->addr = ntohl (prefs.my_ip);
-   else
+   } else if (prefs.local_ip != 0)
+   {
+      dcc->addr = ntohl (prefs.local_ip);
+   } else
+   {
       dcc->addr = ntohl (SAddr.sin_addr.s_addr);
+   }
 
    fcntl (dcc->sok, F_SETFL, O_NONBLOCK);
    listen (dcc->sok, 1);
@@ -787,6 +820,16 @@ dcc_listen_init (struct DCC *dcc)
    return TRUE;
 }
 
+struct session *dccsess;
+char *dccto; /* lame!! */
+char *dcctbuf;
+
+void
+dcc_send_wild (char *file)
+{
+   dcc_send(dccsess, dcctbuf, dccto, file);
+}
+
 void
 dcc_send (struct session *sess, char *tbuf, char *to, char *file)
 {
@@ -798,6 +841,22 @@ dcc_send (struct session *sess, char *tbuf, char *to, char *file)
    strcpy (dcc->file, file);
    check_homedir (dcc->file);
 
+   if(strchr(dcc->file, '*') || strchr(dcc->file, '?'))
+   {
+	char path[256];
+	char wild[256];
+
+	strcpy(wild, file_part(dcc->file));
+	path_part(dcc->file, path);
+	path[ strlen(path) - 1 ] = 0; /* remove trailing slash */
+	dccsess = sess;
+	dccto = to;
+	dcctbuf = tbuf;
+	for_files(path, wild, dcc_send_wild);
+      dcc_close (dcc, 0, TRUE);
+	return;
+   }
+
    if (stat (dcc->file, &st) != -1)
    {
       if (*file_part (dcc->file))
@@ -809,11 +868,7 @@ dcc_send (struct session *sess, char *tbuf, char *to, char *file)
             dcc->stat = STAT_QUEUED;
             dcc->size = st.st_size;
             dcc->type = TYPE_SEND;
-#ifdef __EMX__
-            dcc->fp = open (dcc->file, O_BINARY | O_RDONLY);
-#else
-            dcc->fp = open (dcc->file, O_RDONLY);
-#endif
+            dcc->fp = open (dcc->file, OFLAGS | O_RDONLY);
             if (dcc->fp != -1)
             {
                if (dcc_listen_init (dcc))
@@ -892,6 +947,23 @@ find_dcc (char *nick, char *file, int type)
    return 0;
 }
 
+/* called when we receive a NICK change from server */
+
+void dcc_change_nick (struct server * serv, char *oldnick, char *newnick)
+{
+   GSList *list = dcc_list;
+   while (list)
+   {
+	struct DCC *dcc = (struct DCC *)list -> data;
+	if (dcc->serv == serv)
+	{
+	   if(!strcasecmp(dcc->nick, oldnick))
+		strcpy(dcc->nick, newnick);
+	}
+	list = list -> next;
+   }
+}
+
 static void
 dcc_abort (struct DCC *dcc)
 {
@@ -981,8 +1053,10 @@ new_dcc (void)
    if (!dcc)
       return 0;
    memset (dcc, 0, sizeof (struct DCC));
+   dcc->perc = 0.00;
    dcc->sok = -1;
    dcc->fp = -1;
+   dcc->wiotag = -1;
    dcc_list = g_slist_append (dcc_list, dcc);
    return (dcc);
 }
@@ -1035,6 +1109,10 @@ dcc_chat (struct session *sess, char *nick)
    strcpy (dcc->nick, nick);
    if (dcc_listen_init (dcc))
    {
+      if (!prefs.noautoopendccchatwindow)
+         open_dcc_chat_window ();
+      else
+         update_dcc_chat_window ();
       snprintf (outbuf, sizeof outbuf, "PRIVMSG %s :\001DCC CHAT chat %lu %ld\001\r\n",
           nick, dcc->addr, dcc->port);
       tcp_send (dcc->serv, outbuf);
@@ -1046,12 +1124,8 @@ dcc_chat (struct session *sess, char *nick)
 static void
 dcc_malformed (struct session *sess, char *nick, char *data)
 {
-   char tbuf[1024];
-   snprintf (tbuf, sizeof tbuf,
-             STARTON " Received a malformed DCC request from %s\n"
-   STARTON " Exact contents: \"%s\"\n",
-             nick, data);
-   PrintText (sess, tbuf);
+   EMIT_SIGNAL (XP_TE_MALFORMED_FROM, sess, nick, NULL, NULL, NULL, 0);
+   EMIT_SIGNAL (XP_TE_MALFORMED_PACKET, sess, data, NULL, NULL, NULL, 0);
 }
 
 void
@@ -1060,7 +1134,7 @@ handle_dcc (struct session *sess, char *outbuf, char *nick, char *word[], char *
    struct DCC *dcc;
    char *type = word[5];
    int port;
-
+/*
    if (!strcasecmp (type, "DRAW"))
    {
       unsigned long addr;
@@ -1073,11 +1147,18 @@ handle_dcc (struct session *sess, char *outbuf, char *nick, char *word[], char *
          dcc_malformed (sess, nick, word_eol[4] + 2);
          return;
       }
+
+	dcc = find_dcc (nick, "", TYPE_DRAWSEND);
+	if(!dcc)
+	   dcc = find_dcc (nick, "", TYPE_DRAWRECV);
+	if (dcc)
+	   dcc_close (dcc, 0, TRUE);
+
       dcc = new_dcc ();
       if (dcc)
       {
-         dcc->pos = atoi (word[6]);  /* x */
-         dcc->size = atoi (word[7]);  /* y */
+         dcc->pos = atoi (word[6]);  // x
+         dcc->size = atoi (word[7]);  // y
          dcc->serv = sess->server;
          dcc->type = TYPE_DRAWRECV;
          dcc->stat = STAT_QUEUED;
@@ -1089,7 +1170,7 @@ handle_dcc (struct session *sess, char *outbuf, char *nick, char *word[], char *
                  NULL, NULL, NULL, 0);
          return;
       }
-   }
+   }*/
    if (!strcasecmp (type, "CHAT"))
    {
       unsigned long addr;
@@ -1103,10 +1184,15 @@ handle_dcc (struct session *sess, char *outbuf, char *nick, char *word[], char *
          dcc_malformed (sess, nick, word_eol[4] + 2);
          return;
       }
+
       dcc = find_dcc (nick, "", TYPE_CHATSEND);
       if (dcc)
          dcc_close (dcc, 0, TRUE);
 
+	dcc = find_dcc (nick, "", TYPE_CHATRECV);
+	if (dcc)
+	   dcc_close (dcc, 0, TRUE);
+
       dcc = new_dcc ();
       if (dcc)
       {
@@ -1119,6 +1205,10 @@ handle_dcc (struct session *sess, char *outbuf, char *nick, char *word[], char *
 
          EMIT_SIGNAL (XP_TE_DCCCHATOFFER, sess->server->front_session, nick,
                  NULL, NULL, NULL, 0);
+         if (!prefs.noautoopendccchatwindow)
+            open_dcc_chat_window ();
+         else
+            update_dcc_chat_window ();
          if (prefs.autodccchat)
             dcc_connect (0, dcc);
          return;
@@ -1205,9 +1295,14 @@ handle_dcc (struct session *sess, char *outbuf, char *nick, char *word[], char *
             update_dcc_recv_window ();
          if (prefs.autodccsend)
          {
-            dcc->resumable = 0;
-            dcc->pos = 0;
-            dcc_connect (sess, dcc);
+		if (prefs.autoresume && dcc->resumable)
+		{
+		   dcc_resume (dcc);
+		} else {
+         	   dcc->resumable = 0;
+         	   dcc->pos = 0;
+         	   dcc_connect (sess, dcc);
+		}
          }
       }
       sprintf (outbuf, "%ld", size);
@@ -1226,8 +1321,8 @@ dcc_show_list (struct session *sess, char *outbuf)
    GSList *list = dcc_list;
 
    PrintText (sess,
-              "\00308,02 Type  To/From    Status  Size    Pos     File   \003\n"
-              "\002\00303-------------------------------------------------\002\003\n"
+              "\00308,02 Type  To/From    Status  Size    Pos     File    \017\n"
+              "\002\00309--------------------------------------------------\017\n"
       );
    while (list)
    {
@@ -1235,7 +1330,7 @@ dcc_show_list (struct session *sess, char *outbuf)
       i++;
       snprintf (outbuf, 255, " %-5.5s %-10.10s %-7.7s %-7ld %-7ld %s\n",
                 dcctypes[(int) dcc->type], dcc->nick,
-                dccstat[(int) dcc->stat].name, dcc->size, dcc->pos, dcc->file);
+                dccstat[(int) dcc->stat].name, dcc->size, dcc->pos, file_part (dcc->file));
       PrintText (sess, outbuf);
       list = list->next;
    }
@@ -1255,10 +1350,11 @@ update_dcc_recv_window (void)
 {
    struct DCC *dcc;
    GSList *list = dcc_list;
-   gchar *new[1][7];
+   gchar *nnew[8];
    char size[16];
    char pos[16];
    char cps[16];
+   char perc[14];
 #ifdef USE_GNOME
    char file[512];
 #endif
@@ -1271,38 +1367,39 @@ update_dcc_recv_window (void)
    selrow = gtkutil_clist_selection (dccrwin.list);
 
    gtk_clist_clear ((GtkCList *) dccrwin.list);
-   new[0][2] = size;
-   new[0][3] = pos;
-   new[0][4] = cps;
+   nnew[2] = size;
+   nnew[3] = pos;
+   nnew[4] = perc;
+   nnew[5] = cps;
    while (list)
    {
       dcc = (struct DCC *) list->data;
       if (dcc->type == TYPE_RECV)
       {
-         new[0][0] = dccstat[(int) dcc->stat].name;
-         new[0][1] = dcc->file;
-         new[0][5] = dcc->nick;
+         nnew[0] = dccstat[(int) dcc->stat].name;
+         nnew[1] = dcc->file;
+         nnew[6] = dcc->nick;
 #ifdef USE_GNOME
          if (dcc->stat == STAT_DONE)
          {
             snprintf (file, sizeof (file), "%s/%s", prefs.dccdir, dcc->file);
-            new[0][6] = (char *) gnome_mime_type_of_file (file);
+            nnew[7] = (char *) gnome_mime_type_of_file (file);
          } else
-            new[0][6] = "";
+            nnew[7] = "";
 #endif
          sprintf (size, "%lu", dcc->size);
          if (dcc->stat == STAT_QUEUED)
             sprintf (pos, "%lu", dcc->resumable);
          else
             sprintf (pos, "%lu", dcc->pos);
-         sprintf (cps, "%lu", dcc->cps);
-         row = gtk_clist_append (GTK_CLIST (dccrwin.list), new[0]);
+         snprintf (cps, sizeof (cps), "%lu", dcc->cps);
+         snprintf (perc, sizeof (perc), "%.0f%%", dcc->perc);
+         row = gtk_clist_append (GTK_CLIST (dccrwin.list), nnew);
          gtk_clist_set_row_data (GTK_CLIST (dccrwin.list), row, (gpointer) dcc);
-         gtk_clist_set_foreground
-            (
-         GTK_CLIST (dccrwin.list), row,
-               colors + dccstat[(int) dcc->stat].color
-            );
+	   if (dccstat[(int) dcc->stat].color != 1)
+      	gtk_clist_set_foreground(
+      	   GTK_CLIST (dccrwin.list), row,
+               colors + dccstat[(int) dcc->stat].color);
       }
       list = list->next;
    }
@@ -1402,10 +1499,10 @@ open_dcc_recv_window (void)
    GtkWidget *vbox, *bbox;
 #ifdef USE_GNOME
    static gchar *titles[] =
-   {"Status", "File", "Size", "Position", "CPS", "From", "MIME Type"};
+   {"Status", "File", "Size", "Position", "%", "CPS", "From", "MIME Type"};
 #else
    static gchar *titles[] =
-   {"Status", "File", "Size", "Position", "CPS", "From"};
+   {"Status", "File", "Size", "Position", "%", "CPS", "From"};
 #endif
 
    if (dccrwin.window)
@@ -1422,11 +1519,11 @@ open_dcc_recv_window (void)
    gtk_widget_show (vbox);
 
 #ifdef USE_GNOME
-   dccrwin.list = gtkutil_clist_new (7, titles, vbox, GTK_POLICY_ALWAYS,
+   dccrwin.list = gtkutil_clist_new (8, titles, vbox, GTK_POLICY_ALWAYS,
                  recv_row_selected, 0,
           0, 0, GTK_SELECTION_SINGLE);
 #else
-   dccrwin.list = gtkutil_clist_new (6, titles, vbox, GTK_POLICY_ALWAYS,
+   dccrwin.list = gtkutil_clist_new (7, titles, vbox, GTK_POLICY_ALWAYS,
                  recv_row_selected, 0,
           0, 0, GTK_SELECTION_SINGLE);
 #endif
@@ -1434,20 +1531,29 @@ open_dcc_recv_window (void)
    gtk_clist_set_column_width (GTK_CLIST (dccrwin.list), 1, 100);
    gtk_clist_set_column_width (GTK_CLIST (dccrwin.list), 2, 50);
    gtk_clist_set_column_width (GTK_CLIST (dccrwin.list), 3, 50);
-   gtk_clist_set_column_width (GTK_CLIST (dccrwin.list), 4, 50);
-   gtk_clist_set_column_width (GTK_CLIST (dccrwin.list), 5, 60);
+   gtk_clist_set_column_width (GTK_CLIST (dccrwin.list), 4, 30);
+   gtk_clist_set_column_width (GTK_CLIST (dccrwin.list), 5, 50);
+   gtk_clist_set_column_width (GTK_CLIST (dccrwin.list), 6, 60);
+
+   /*gtk_clist_set_column_justification (GTK_CLIST (dccrwin.list), 0, GTK_JUSTIFY_LEFT);
+      gtk_clist_set_column_justification (GTK_CLIST (dccrwin.list), 1, GTK_JUSTIFY_LEFT);
+      gtk_clist_set_column_justification (GTK_CLIST (dccrwin.list), 2, GTK_JUSTIFY_LEFT);
+      gtk_clist_set_column_justification (GTK_CLIST (dccrwin.list), 3, GTK_JUSTIFY_LEFT); */
+   gtk_clist_set_column_justification (GTK_CLIST (dccrwin.list), 4, GTK_JUSTIFY_CENTER);
+   /*gtk_clist_set_column_justification (GTK_CLIST (dccrwin.list), 5, GTK_JUSTIFY_LEFT);
+      gtk_clist_set_column_justification (GTK_CLIST (dccrwin.list), 6, GTK_JUSTIFY_LEFT); */
 
    bbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 0);
    gtk_widget_show (bbox);
 
-   gtkutil_button_new ("Accept", 0, 0, accept_clicked, 0, bbox);
+   gtkutil_button (dccrwin.window, 0, "Accept", accept_clicked, 0, bbox);
+   
+   gtkutil_button (dccrwin.window, 0, "Resume", resume_clicked, 0, bbox);
 
-   gtkutil_button_new ("Resume", 0, 0, resume_clicked, 0, bbox);
+   gtkutil_button (dccrwin.window, 0, "Abort", abort_clicked, 0, bbox);
 
-   gtkutil_button_new ("Abort", 0, 0, abort_clicked, 0, bbox);
-
-   gtkutil_button_new ("Info", 0, 0, info_clicked, 0, bbox);
+   gtkutil_button (dccrwin.window, 0, "Info", info_clicked, 0, bbox);
 
    gtk_widget_show (dccrwin.window);
    update_dcc_recv_window ();
@@ -1465,11 +1571,12 @@ update_dcc_send_window (void)
 {
    struct DCC *dcc;
    GSList *list = dcc_list;
-   gchar *new[1][7];
+   gchar *nnew[8];
    char size[14];
    char pos[14];
    char cps[14];
    char ack[14];
+   char perc[14];
    gint row;
    int selrow;
 
@@ -1479,29 +1586,31 @@ update_dcc_send_window (void)
    selrow = gtkutil_clist_selection (dccswin.list);
 
    gtk_clist_clear ((GtkCList *) dccswin.list);
-   new[0][2] = size;
-   new[0][3] = pos;
-   new[0][4] = ack;
-   new[0][5] = cps;
+   nnew[2] = size;
+   nnew[3] = pos;
+   nnew[4] = ack;
+   nnew[5] = perc;
+   nnew[6] = cps;
    while (list)
    {
       dcc = (struct DCC *) list->data;
       if (dcc->type == TYPE_SEND)
       {
-         new[0][0] = dccstat[(int) dcc->stat].name;
-         new[0][1] = file_part (dcc->file);
-         new[0][6] = dcc->nick;
-         sprintf (size, "%lu", dcc->size);
-         sprintf (pos, "%lu", dcc->pos);
-         sprintf (cps, "%lu", dcc->cps);
-         sprintf (ack, "%lu", dcc->ack);
-         row = gtk_clist_append (GTK_CLIST (dccswin.list), new[0]);
+         nnew[0] = dccstat[(int) dcc->stat].name;
+         nnew[1] = file_part (dcc->file);
+         nnew[7] = dcc->nick;
+         snprintf (size, sizeof (size), "%lu", dcc->size);
+         snprintf (pos, sizeof (pos), "%lu", dcc->pos);
+         snprintf (cps, sizeof (cps), "%lu", dcc->cps);
+         snprintf (perc, sizeof (perc), "%.0f%%", dcc->perc);
+         snprintf (ack, sizeof (ack), "%lu", dcc->ack + dcc->resumable);
+         row = gtk_clist_append (GTK_CLIST (dccswin.list), nnew);
          gtk_clist_set_row_data (GTK_CLIST (dccswin.list), row, (gpointer) dcc);
-         gtk_clist_set_foreground
-            (
-         GTK_CLIST (dccswin.list), row,
-               colors + dccstat[(int) dcc->stat].color
-            );
+	   if (dccstat[(int) dcc->stat].color != 1)
+      	gtk_clist_set_foreground
+         	(
+      	   GTK_CLIST (dccswin.list), row,
+               colors + dccstat[(int) dcc->stat].color);
       }
       list = list->next;
    }
@@ -1565,7 +1674,7 @@ open_dcc_send_window (void)
 {
    GtkWidget *vbox, *bbox;
    static gchar *titles[] =
-   {"Status", "File", "Size", "Position", "Ack", "CPS", "To"};
+   {"Status", "File", "Size", "Position", "Ack", "%", "CPS", "To"};
 
    if (dccswin.window)
    {
@@ -1580,7 +1689,7 @@ open_dcc_send_window (void)
    gtk_container_add (GTK_CONTAINER (dccswin.window), vbox);
    gtk_widget_show (vbox);
 
-   dccswin.list = gtkutil_clist_new (7, titles, vbox, GTK_POLICY_ALWAYS,
+   dccswin.list = gtkutil_clist_new (8, titles, vbox, GTK_POLICY_ALWAYS,
                  send_row_selected, 0,
           0, 0, GTK_SELECTION_SINGLE);
    gtk_clist_set_column_width (GTK_CLIST (dccswin.list), 0, 65);
@@ -1588,15 +1697,24 @@ open_dcc_send_window (void)
    gtk_clist_set_column_width (GTK_CLIST (dccswin.list), 2, 50);
    gtk_clist_set_column_width (GTK_CLIST (dccswin.list), 3, 50);
    gtk_clist_set_column_width (GTK_CLIST (dccswin.list), 4, 50);
-   gtk_clist_set_column_width (GTK_CLIST (dccswin.list), 5, 50);
+   gtk_clist_set_column_width (GTK_CLIST (dccswin.list), 5, 30);
+   gtk_clist_set_column_width (GTK_CLIST (dccswin.list), 6, 50);
+
+   /*gtk_clist_set_column_justification (GTK_CLIST (dccswin.list), 0, GTK_JUSTIFY_LEFT);
+      gtk_clist_set_column_justification (GTK_CLIST (dccswin.list), 1, GTK_JUSTIFY_LEFT);
+      gtk_clist_set_column_justification (GTK_CLIST (dccswin.list), 2, GTK_JUSTIFY_LEFT);
+      gtk_clist_set_column_justification (GTK_CLIST (dccswin.list), 3, GTK_JUSTIFY_LEFT);
+      gtk_clist_set_column_justification (GTK_CLIST (dccswin.list), 4, GTK_JUSTIFY_LEFT); */
+   gtk_clist_set_column_justification (GTK_CLIST (dccswin.list), 5, GTK_JUSTIFY_CENTER);
+   /*gtk_clist_set_column_justification (GTK_CLIST (dccswin.list), 6, GTK_JUSTIFY_LEFT); */
 
    bbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 0);
    gtk_widget_show (bbox);
 
-   gtkutil_button_new ("Abort", 0, 0, abort_send_clicked, 0, bbox);
+   gtkutil_button (dccswin.window, 0, "Abort", abort_send_clicked, 0, bbox);
 
-   gtkutil_button_new ("Info", 0, 0, info_send_clicked, 0, bbox);
+   gtkutil_button (dccswin.window, 0, "Info", info_send_clicked, 0, bbox);
 
    gtk_widget_show (dccswin.window);
    update_dcc_send_window ();
@@ -1605,6 +1723,31 @@ open_dcc_send_window (void)
 
 /* DCC CHAT GUIs BELOW */
 
+static void
+accept_chat_clicked (GtkWidget * wid, gpointer none)
+{
+   int row;
+   struct DCC *dcc;
+
+   row = gtkutil_clist_selection (dcccwin.list);
+   if (row != -1)
+   {
+      dcc = gtk_clist_get_row_data (GTK_CLIST (dcccwin.list), row);
+      gtk_clist_unselect_row (GTK_CLIST (dcccwin.list), row, 0);
+      switch (dcc->stat)
+      {
+      case STAT_QUEUED:
+	   if(dcc->type == TYPE_CHATRECV)
+      	dcc_connect (0, dcc);
+         break;
+      case STAT_DONE:
+      case STAT_FAILED:
+      case STAT_ABORTED:
+         dcc_abort (dcc);
+         break;
+	}
+   }
+}
 
 static void
 abort_chat_clicked (GtkWidget * wid, gpointer none)
@@ -1620,12 +1763,20 @@ abort_chat_clicked (GtkWidget * wid, gpointer none)
    }
 }
 
+static void
+chat_row_selected (GtkWidget * clist, gint row, gint column,
+                GdkEventButton * even)
+{
+   if (even && even->type == GDK_2BUTTON_PRESS)
+	accept_chat_clicked (0, 0);
+}
+
 static void
 update_dcc_chat_window (void)
 {
    struct DCC *dcc;
    GSList *list = dcc_list;
-   gchar *new[1][5];
+   gchar *nnew[5];
    char pos[14];
    char siz[14];
    gint row;
@@ -1637,19 +1788,19 @@ update_dcc_chat_window (void)
    selrow = gtkutil_clist_selection (dcccwin.list);
 
    gtk_clist_clear ((GtkCList *) dcccwin.list);
-   new[0][2] = pos;
-   new[0][3] = siz;
+   nnew[2] = pos;
+   nnew[3] = siz;
    while (list)
    {
       dcc = (struct DCC *) list->data;
       if ((dcc->type == TYPE_CHATSEND || dcc->type == TYPE_CHATRECV))
       {
-         new[0][0] = dccstat[(int) dcc->stat].name;
-         new[0][1] = dcc->nick;
+         nnew[0] = dccstat[(int) dcc->stat].name;
+         nnew[1] = dcc->nick;
          sprintf (pos, "%lu", dcc->pos);
          sprintf (siz, "%lu", dcc->size);
-         new[0][4] = ctime (&dcc->starttime);
-         row = gtk_clist_append (GTK_CLIST (dcccwin.list), new[0]);
+         nnew[4] = ctime (&dcc->starttime);
+         row = gtk_clist_append (GTK_CLIST (dcccwin.list), nnew);
          gtk_clist_set_row_data (GTK_CLIST (dcccwin.list), row, (gpointer) dcc);
       }
       list = list->next;
@@ -1686,8 +1837,8 @@ open_dcc_chat_window (void)
    gtk_widget_show (vbox);
 
    dcccwin.list = gtkutil_clist_new (5, titles, vbox, GTK_POLICY_ALWAYS,
-                                 0, 0,
-          0, 0, GTK_SELECTION_BROWSE);
+                              	 chat_row_selected, 0,
+                                     0, 0, GTK_SELECTION_BROWSE);
    gtk_clist_set_column_width (GTK_CLIST (dcccwin.list), 0, 65);
    gtk_clist_set_column_width (GTK_CLIST (dcccwin.list), 1, 100);
    gtk_clist_set_column_width (GTK_CLIST (dcccwin.list), 2, 65);
@@ -1697,8 +1848,11 @@ open_dcc_chat_window (void)
    gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 0);
    gtk_widget_show (bbox);
 
-   gtkutil_button_new ("Abort", 0, 0, abort_chat_clicked, 0, bbox);
+   gtkutil_button (dcccwin.window, 0, "Accept", accept_chat_clicked, 0, bbox);
+   gtkutil_button (dcccwin.window, 0, "Abort", abort_chat_clicked, 0, bbox);
 
    gtk_widget_show (dcccwin.window);
    update_dcc_chat_window ();
 }
+
+
