@@ -16,12 +16,13 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
 
+#define SKIPGNOME
 #include "style.h"
 #include "xchat.h"
 #include "util.h"
-#undef USE_GNOME
-
+#include "ignore.h"
 #include "plugin.h"
+#include "userlist.h"
 #include <sys/utsname.h>
 #include <sys/time.h>
 #include <unistd.h>
@@ -44,6 +45,7 @@ void save_away_message (struct server *serv, char *nick, char *msg);
 
 /* text.c */
 
+extern unsigned char *strip_color (unsigned char *text);
 extern void PrintText (struct session *sess, char *text);
 extern void setup_logging (struct session *sess);
 
@@ -70,24 +72,15 @@ extern void add_rawlog (struct server *serv, char *text, int outbound);
 
 /* dcc.c */
 
+extern void dcc_change_nick (struct server * serv, char *oldnick, char *newnick);
 extern void handle_dcc (struct session *sess, char *outbuf, char *nick, char *word[], char *word_eol[]);
 
 /* dialog.c */
 
 extern void dialog_change_nick (struct server *serv, char *nick, char *newnick);
 
-/* userlist.c */
+/* maingui.c */
 
-extern struct user *find_name (struct session *sess, char *name);
-extern void update_user_list (struct session *sess);
-extern void clear_user_list (struct session *sess);
-extern void sort_namelist (struct session *sess);
-extern void voice_name (struct session *sess, char *name);
-extern void devoice_name (struct session *sess, char *name);
-extern void op_name (struct session *sess, char *name);
-extern void deop_name (struct session *sess, char *name);
-extern void add_name (struct session *sess, char *name);
-extern int sub_name (struct session *sess, char *name);
 extern void change_nick (struct session *sess, char *oldnick, char *newnick);
 extern void gui_change_nick (struct server *serv, char *newnick);
 extern int module_inbound (struct session *sess, struct server *serv, char *buf);
@@ -138,8 +131,10 @@ extern struct session *current_tab;
 #define find_word_to_end(a, b) word_eol[b]
 #define find_word(a, b) word[b]
 
-char chan_flags[] =
-{'t', 'n', 's', 'i', 'p', 'm', 'l', 'k'};
+char chan_flags[] = {'t', 'n', 's', 'i', 'p', 'm', 'l', 'k'};
+
+void channel_msg (struct server *serv, char *outbuf, char *chan, char *from, char *text, char fromme);
+
 
 struct session *
 find_unused_session (struct server *serv)
@@ -216,8 +211,8 @@ private_msg (struct server *serv, char *tbuf, char *from, char *ip, char *text)
    {
       if (!sess)
          sess = new_session (serv, from);  /* Create a dialog session */
-
-      EMIT_SIGNAL (XP_TE_DPRIVMSG, sess, from, text, NULL, NULL, 0);
+	channel_msg(serv, tbuf, from, from, text, FALSE);
+      /*EMIT_SIGNAL (XP_TE_DPRIVMSG, sess, from, text, NULL, NULL, 0);*/
       if (ip && ip[0])
          gtk_entry_set_text (GTK_ENTRY (sess->topicgad), ip);
       return;
@@ -250,12 +245,12 @@ channel_action (struct session *sess, char *tbuf, char *chan, char *from, char *
          sprintf (tbuf, from);
    }
 
-   if (chan[0] == '#' || chan[0] == '&' || fromme)
+   if (chan[0] == '#' || chan[0] == '&' || chan[0] == '+' || fromme)
       sess = find_session_from_channel (chan, sess->server);
    else                         /* it's a private action! */
       sess = find_session_from_channel (from, sess->server);
 
-   if (!sess && chan[0] != '#' && chan[0] != '&' && prefs.autodialog)
+   if (!sess && chan[0] != '#' && chan[0] != '&' && chan[0] != '+' && prefs.autodialog)
       sess = new_session (def->server, from);
 
    if (sess)
@@ -267,12 +262,12 @@ channel_action (struct session *sess, char *tbuf, char *chan, char *from, char *
 int
 SearchNick (char *text, char *nicks)
 {
-   char S[50];
+   char S[64];
    char *n;
 
    if (nicks == NULL)
       return 0;
-   strcpy (S, nicks);
+   strncpy (S, nicks, 63);
    n = strtok (S, ",");
    while (n != NULL)
    {
@@ -289,6 +284,7 @@ channel_msg_with_tabnicks (struct server *serv, char *outbuf, char *chan, char *
    char *real_outbuf = outbuf, *po;
    struct session *sess;
    int i, len;
+   int highlight = FALSE;
 
    sess = find_session_from_channel (chan, serv);
    if (!sess)
@@ -314,26 +310,28 @@ channel_msg_with_tabnicks (struct server *serv, char *outbuf, char *chan, char *
          sprintf (outbuf, CHANNEL_MSG_FROM_USER, from, text);
    } else
    {
+      if (nocasestrstr (text, serv->nick) || SearchNick (text, prefs.bluestring))
+      {
+         if (EMIT_SIGNAL (XP_HIGHLIGHT, sess, chan, from, text, NULL, fromme) == 1)
+            return;
+         highlight = TRUE;
+#ifdef USE_PANEL
+         if (sess->panel_button)
+            gtk_widget_set_style (GTK_BIN (sess->panel_button)->child, bluetab_style);
+#endif
+         if (sess != current_tab && sess->is_tab)
+         {
+            sess->nick_said = TRUE;
+            gtk_widget_set_style (sess->changad, bluetab_style);
+         }
+      }
       if (prefs.colorednicks)
          sprintf (outbuf, CHANNEL_MSG_COLOR, color_of (from), from, text);
       else
       {
-         if (nocasestrstr (text, serv->nick) || SearchNick (text, prefs.bluestring))
-         {
-            if (EMIT_SIGNAL (XP_HIGHLIGHT, sess, chan, from, text, NULL, fromme) == 1)
-               return;
-
+         if (highlight)
             sprintf (outbuf, CHANNEL_MSG_WITH_NICK, prefs.bt_color, from, text);
-#ifdef USE_PANEL
-            if (sess->panel_button)
-               gtk_widget_set_style (GTK_BIN (sess->panel_button)->child, bluetab_style);
-#endif
-            if (sess != current_tab && sess->is_tab)
-            {
-               sess->nick_said = TRUE;
-               gtk_widget_set_style (sess->changad, bluetab_style);
-            }
-         } else
+         else
             sprintf (outbuf, CHANNEL_MSG, from, text);
       }
    }
@@ -351,6 +349,7 @@ channel_msg_normal (struct server *serv, char *outbuf, char *chan, char *from, c
 {
    char *real_outbuf = outbuf;
    struct session *sess;
+   int highlight = FALSE;
 
    sess = find_session_from_channel (chan, serv);
    if (!sess)
@@ -380,26 +379,29 @@ channel_msg_normal (struct server *serv, char *outbuf, char *chan, char *from, c
          sprintf (outbuf, from);
    } else
    {
+
+      if (nocasestrstr (text, serv->nick) || SearchNick (text, prefs.bluestring))
+      {
+         if (EMIT_SIGNAL (XP_HIGHLIGHT, sess, chan, from, text, NULL, fromme) == 1)
+            return;
+         highlight = TRUE;
+#ifdef USE_PANEL
+         if (sess->panel_button)
+            gtk_widget_set_style (GTK_BIN (sess->panel_button)->child, bluetab_style);
+#endif
+         if (sess != current_tab && sess->is_tab)
+         {
+            sess->nick_said = TRUE;
+            gtk_widget_set_style (sess->changad, bluetab_style);
+         }
+      }
       if (prefs.colorednicks)
          sprintf (outbuf, "\003%d%s\017", color_of (from), from);
       else
       {
-         if (nocasestrstr (text, serv->nick) || SearchNick (text, prefs.bluestring))
-         {
-            if (EMIT_SIGNAL (XP_HIGHLIGHT, sess, chan, from, text, NULL, fromme) == 1)
-               return;
-
+         if (highlight)
             sprintf (outbuf, "\002\003%d%s\002\017", prefs.bt_color, from);
-#ifdef USE_PANEL
-            if (sess->panel_button)
-               gtk_widget_set_style (GTK_BIN (sess->panel_button)->child, bluetab_style);
-#endif
-            if (sess != current_tab && sess->is_tab)
-            {
-               sess->nick_said = TRUE;
-               gtk_widget_set_style (sess->changad, bluetab_style);
-            }
-         } else
+         else
             sprintf (outbuf, from);
       }
    }
@@ -465,6 +467,7 @@ user_new_nick (struct server *serv, char *outbuf, char *nick, char *newnick, int
    }
 
    dialog_change_nick (serv, nick, newnick);
+   dcc_change_nick (serv, nick, newnick);
 
    if (me)
       gui_change_nick (serv, newnick);
@@ -481,6 +484,12 @@ you_joined (struct server *serv, char *outbuf, char *chan, char *nick, char *ip)
       tcp_send (sess->server, outbuf);
       clear_user_list (sess);
       EMIT_SIGNAL (XP_TE_UJOIN, sess, nick, chan, ip, NULL, 0);
+      if (!prefs.nouserhost)
+      {
+         sprintf (outbuf, "WHO %s\r\n", sess->channel);
+         tcp_send (serv, outbuf);
+         sess->doing_who = TRUE;
+      }
    }
 }
 
@@ -518,60 +527,52 @@ you_parted (struct server *serv, char *tbuf, char *chan)
 }
 
 void
-names_list (struct server *serv, char *tbuf, char *buf)
+names_list (struct server *serv, char *tbuf, char *chan, char *names)
 {
-   char *po;
-   buf += 2;
-   po = strchr (buf, ' ');
-   if (po)
+   struct session *sess;
+   char name[64];
+   int pos = 0;
+
+   sess = find_session_from_channel (chan, serv);
+   if (!sess)
    {
-      struct session *sess;
-      char name[64];
-      int pos = 0;
-
-      po[0] = 0;
-
-      sess = find_session_from_channel (buf, serv);
-      if (!sess)
-      {
-         EMIT_SIGNAL (XP_TE_USERSONCHAN, serv->front_session, buf, po + 2, NULL,
+	EMIT_SIGNAL (XP_TE_USERSONCHAN, serv->front_session, chan, names, NULL,
                       NULL, 0);
+   	return;
+   }
+   if (!sess->ignore_names)
+      EMIT_SIGNAL (XP_TE_USERSONCHAN, sess, chan, names, NULL, NULL, 0);
+
+   if (sess->end_of_names)
+   {
+      sess->end_of_names = FALSE;
+      clear_user_list (sess);
+   }
+
+   gtk_clist_freeze (GTK_CLIST (sess->namelistgad));
+
+   while (1)
+   {
+      switch (*names)
+      {
+      case 0:
+         name[pos] = 0;
+         if (pos != 0)
+            add_name (sess, name, 0);
+         gtk_clist_thaw (GTK_CLIST (sess->namelistgad));
          return;
+      case ' ':
+         name[pos] = 0;
+         pos = 0;
+         add_name (sess, name, 0);
+         break;
+      default:
+         name[pos] = *names;
+         pos++;
+         if (pos == 62)
+            pos = 61;
       }
-      if (!sess->ignore_names)
-         EMIT_SIGNAL (XP_TE_USERSONCHAN, sess, buf, po + 2, NULL, NULL, 0);
-
-      if (sess->end_of_names)
-      {
-         sess->end_of_names = FALSE;
-         clear_user_list (sess);
-      }
-      buf = po + 1;
-
-      gtk_clist_freeze (GTK_CLIST (sess->namelistgad));
-      while (1)
-      {
-         buf++;
-         switch (*buf)
-         {
-         case 0:
-            name[pos] = 0;
-            if (pos != 0)
-               add_name (sess, name);
-            gtk_clist_thaw (GTK_CLIST (sess->namelistgad));
-            return;
-         case ' ':
-            name[pos] = 0;
-            pos = 0;
-            add_name (sess, name);
-            break;
-         default:
-            name[pos] = *buf;
-            pos++;
-            if (pos == 51)
-               pos = 50;
-         }
-      }
+	names++;
    }
 }
 
@@ -586,7 +587,9 @@ topic (struct server *serv, char *tbuf, char *buf)
       sess = find_session_from_channel (buf, serv);
       if (sess)
       {
-         gtk_entry_set_text (GTK_ENTRY (sess->topicgad), po + 2);
+	   char *new_topic = strip_color(po + 2);
+         gtk_entry_set_text (GTK_ENTRY (sess->topicgad), new_topic);
+	   free (new_topic);
          EMIT_SIGNAL (XP_TE_TOPIC, sess, buf, po + 2, NULL, NULL, 0);
       }
    }
@@ -613,8 +616,9 @@ user_joined (struct server *serv, char *outbuf, char *chan, char *user, char *ip
 
    if (sess)
    {
-      EMIT_SIGNAL (XP_TE_JOIN, sess, user, chan, ip, NULL, 0);
-      add_name (sess, user);
+      if (!(GTK_TOGGLE_BUTTON (sess->confbutton)->active))
+         EMIT_SIGNAL (XP_TE_JOIN, sess, user, chan, ip, NULL, 0);
+      add_name (sess, user, ip);
    }
 }
 
@@ -635,7 +639,8 @@ user_parted (struct server *serv, char *chan, char *user, char *ip)
    struct session *sess = find_session_from_channel (chan, serv);
    if (sess)
    {
-      EMIT_SIGNAL (XP_TE_PART, sess, user, ip, chan, NULL, 0);
+      if (!(GTK_TOGGLE_BUTTON (sess->confbutton)->active))
+         EMIT_SIGNAL (XP_TE_PART, sess, user, ip, chan, NULL, 0);
       sub_name (sess, user);
    }
 }
@@ -667,7 +672,7 @@ topic_nametime (struct server *serv, char *tbuf, char *chan, char *nick, char *d
 }
 
 void
-set_server_name (char *tbuf, struct server *serv, char *name)
+set_server_name (struct server *serv, char *name)
 {
    GSList *list = sess_list;
    struct session *sess;
@@ -679,6 +684,11 @@ set_server_name (char *tbuf, struct server *serv, char *name)
          gui_set_title (sess);
       list = list->next;
    }
+   if (serv->front_session->is_server)
+   {
+	strcpy (serv->front_session->channel, name);
+	set_channel (serv->front_session);
+   }
 }
 
 void
@@ -697,10 +707,10 @@ handle_ctcp (struct session *sess, char *outbuf, char *to, char *nick, char *msg
 
          mhz = get_mhz ();
          if (mhz)
-            sprintf (outbuf, "NOTICE %s :\001VERSION X-Chat " VERSION " By ZED - %s %s [%s/%dMHz]: http://xchat.linuxpower.org\001\r\n",
+            sprintf (outbuf, "NOTICE %s :\001VERSION X-Chat " VERSION " By ZED - %s %s [%s/%dMHz]: http://xchat.org\001\r\n",
                      nick, un.sysname, un.release, un.machine, mhz);
          else
-            sprintf (outbuf, "NOTICE %s :\001VERSION X-Chat " VERSION " By ZED - %s %s [%s]: http://xchat.linuxpower.org\001\r\n",
+            sprintf (outbuf, "NOTICE %s :\001VERSION X-Chat " VERSION " By ZED - %s %s [%s]: http://xchat.org\001\r\n",
                      nick, un.sysname, un.release, un.machine);
 
          tcp_send (sess->server, outbuf);
@@ -741,15 +751,13 @@ handle_ctcp (struct session *sess, char *outbuf, char *to, char *nick, char *msg
    po = strchr (msg, '\001');
    if (po)
       po[0] = 0;
-   if (*to != '#' && *to != '&')
+   if (*to != '#' && *to != '&' && *to != '+')
    {
-      /*     sprintf(outbuf, STARTON" Received a CTCP %s from %s\n", msg, nick); */
       EMIT_SIGNAL (XP_TE_CTCPGEN, sess->server->front_session, msg, nick, NULL, NULL, 0);
       return;
    } else
    {
       struct session *chansess = find_session_from_channel (to, sess->server);
-      /*     sprintf(outbuf, STARTON" Received a CTCP %s from %s (to %s)\n", msg, nick, to); */
       if (chansess)
       {
          EMIT_SIGNAL (XP_TE_CTCPGENC, chansess, msg, nick, to, NULL, 0);
@@ -771,8 +779,11 @@ user_quit (struct server *serv, char *outbuf, char *nick, char *reason)
       if (sess->server == serv)
       {
          if (sub_name (sess, nick))
-            EMIT_SIGNAL (XP_TE_QUIT, sess, nick, reason,
+         {
+            if (!(GTK_TOGGLE_BUTTON (sess->confbutton)->active))
+               EMIT_SIGNAL (XP_TE_QUIT, sess, nick, reason,
                        NULL, NULL, 0);
+         }
       }
       list = list->next;
    }
@@ -791,10 +802,8 @@ got_ping_reply (struct session *sess, char *outbuf,
    dif = nowtim - tim;
 
    if (atol (timestring) == 0)
-   {
-      sprintf (outbuf, STARTON " Received ping reply from %s.\n", from);
-      PrintText (sess, outbuf);
-   } else
+      EMIT_SIGNAL (XP_TE_PINGREP, sess, from, "?", NULL, NULL, 0);
+   else
    {
       sprintf (outbuf, "%ld.%ld", dif / 1000000, (dif / 100000) % 10);
       EMIT_SIGNAL (XP_TE_PINGREP, sess, from, outbuf, NULL, NULL, 0);
@@ -807,7 +816,7 @@ notice (struct server *serv, char *outbuf, char *to, char *nick, char *msg, char
    char *po;
    struct session *sess = 0;
 
-   if (to[0] == '#' || to[0] == '&')
+   if (to[0] == '#' || to[0] == '&' || to[0] == '+')
       sess = find_session_from_channel (to, serv);
    if (!sess)
    {
@@ -968,7 +977,7 @@ channel_mode (struct server *serv, char *outbuf, char *chan, char *nick, char si
          case 'o':
             if (EMIT_SIGNAL (XP_CHANOP, sess, chan, nick, extra, NULL, 0) == 1)
                return;
-            op_name (sess, extra);
+            ul_op_name (sess, extra);
             if (!quiet)
                EMIT_SIGNAL (XP_TE_CHANOP, sess, nick, extra, NULL, NULL, 0);
             return;
@@ -1039,10 +1048,8 @@ channel_mode (struct server *serv, char *outbuf, char *chan, char *nick, char si
          tbuf[1][0] = mode;
          tbuf[1][1] = 0;
          if (extra && *extra)
-            /*    sprintf(outbuf, STARTON" %s sets mode %c%c %s\n", nick, sign, mode, extra); */
             EMIT_SIGNAL (XP_TE_CHANMODEGEN, sess, nick, tbuf[0], tbuf[1], extra, 0);
          else
-            /*    sprintf(outbuf, STARTON" %s sets mode %c%c %s\n", nick, sign, mode, chan); */
             EMIT_SIGNAL (XP_TE_CHANMODEGEN, sess, nick, tbuf[0], tbuf[1], extra, 0);
       }
       update_channelmode_buttons (sess, mode, sign);
@@ -1209,16 +1216,16 @@ next_nick (struct session *sess, char *outbuf, char *nick)
 }
 
 void
-do_dns (struct session *sess, char *tbuf, char *host)
+do_dns (struct session *sess, char *tbuf, char *nick, char *host)
 {
    char *po;
 
-   po = strchr (host, '@');
+   po = strrchr (host, '@');
    if (po)
-   {
-      sprintf (tbuf, "/exec nslookup %s", po + 1);
-      handle_command (tbuf, sess, 0, 0);
-   }
+      host = po + 1;
+   EMIT_SIGNAL (XP_TE_RESOLVINGUSER, sess, nick, host, NULL, NULL, 0);
+   sprintf (tbuf, "/exec %s %s", prefs.dnsprogram, host);
+   handle_command (tbuf, sess, 0, 0);
 }
 
 /* process_line() */
@@ -1236,12 +1243,12 @@ process_line (struct session *sess, struct server *serv, char *buf)
 
    add_rawlog (serv, buf, FALSE);
    url_checkurl (buf);
-   command_level = 0;
+/*   command_level = 0;*/
 
    process_data_init (pdibuf, buf + 1, word, word_eol);
 
 #ifdef USE_PERL
-   if (*buf == ':' && (word[3][0] == '#' || word[3][0] == '&'))
+   if (*buf == ':' && (word[3][0] == '#' || word[3][0] == '&' || word[3][0] == '+'))
    {
       sess = find_session_from_channel (word[3], serv);
       if (!sess)
@@ -1283,7 +1290,7 @@ process_line (struct session *sess, struct server *serv, char *buf)
       }
       if (!strncmp (buf, "ERROR", 5))
       {
-         EMIT_SIGNAL (XP_TE_SERVERERROR, serv->front_session, buf + 7, NULL,
+         EMIT_SIGNAL (XP_TE_SERVERERROR, sess, buf + 7, NULL,
                       NULL, NULL, 0);
          return 0;
       }
@@ -1307,31 +1314,25 @@ process_line (struct session *sess, struct server *serv, char *buf)
             {
             case 1:
                user_new_nick (serv, outbuf, serv->nick, word[3], TRUE);
-               set_server_name (outbuf, serv, pdibuf);
+               set_server_name (serv, pdibuf);
                goto def;
             case 301:
                handle_away (serv, outbuf, find_word (pdibuf, 4),
                             find_word_to_end (buf, 5) + 1);
                break;
-            case 302:          /* USERHOST */
-               if (!sess->server->doing_dns)
-                  goto def;
-               sess->server->doing_dns = FALSE;
-               do_dns (sess, outbuf, word[4]);
-               break;
             case 303:
                word[4]++;
-               notify_markonline (serv->front_session, outbuf, word);
+               notify_markonline (sess, outbuf, word);
                break;
             case 312:
-               EMIT_SIGNAL (XP_TE_WHOIS3, serv->front_session, word[4],
+               EMIT_SIGNAL (XP_TE_WHOIS3, sess, word[4],
                             word_eol[5], NULL, NULL, 0);
                break;
             case 311:
                serv->inside_whois = 1;
                /* FALL THROUGH */
             case 314:
-               EMIT_SIGNAL (XP_TE_WHOIS1, serv->front_session,
+               EMIT_SIGNAL (XP_TE_WHOIS1, sess,
                             find_word (pdibuf, 4), find_word (pdibuf, 5),
                             find_word (pdibuf, 6), find_word_to_end (buf, 8) + 1, 0);
                break;
@@ -1372,7 +1373,7 @@ process_line (struct session *sess, struct server *serv, char *buf)
                                      find_word_to_end (buf, 6) + 1);
                } else
                {
-                  sprintf (outbuf, "%-16.16s %-7d %s\n",
+                  sprintf (outbuf, "%-16.16s %-7d %s\017\n",
                            find_word (pdibuf, 4),
                            atoi (find_word (pdibuf, 5)),
                            find_word_to_end (buf, 6) + 1);
@@ -1382,6 +1383,7 @@ process_line (struct session *sess, struct server *serv, char *buf)
             case 323:
                if (!sess->server->chanlist_window)
                   goto def;
+               gtk_widget_set_sensitive (sess->server->chanlist_refresh, TRUE);
                break;
             case 324:
                sess = find_session_from_channel (word[4], serv);
@@ -1426,31 +1428,80 @@ process_line (struct session *sess, struct server *serv, char *buf)
                break;
             case 352:          /* WHO */
                if (!serv->skip_next_who)
-                  goto def;
-               if (!strcasecmp (word[8], serv->nick))
                {
-                  struct hostent *HostAddr = gethostbyname (word[6]);
-                  if (HostAddr)
+                  struct session *who_sess;
+
+                  who_sess = find_session_from_channel (word[4], serv);
+                  if (who_sess)
                   {
-                     prefs.my_ip = ((struct in_addr *) HostAddr->h_addr)->s_addr;
-                     EMIT_SIGNAL (XP_TE_FOUNDIP, sess,
-                                  inet_ntoa (*((struct in_addr *) HostAddr->h_addr)), NULL, NULL, NULL, 0);
+                     sprintf (outbuf, "%s@%s", word[5], word[6]);
+                     if (!userlist_add_hostname (who_sess, word[8], outbuf))
+			   {
+				if(!who_sess->doing_who)
+                     	   goto def;
+			   }
+                  } else
+                  {
+                     if (serv->doing_who)
+                        do_dns (sess, outbuf, word[8], word[6]);
+                     else
+                        goto def;
+                  }
+               } else
+               {
+                  if (!strcasecmp (word[8], serv->nick))
+                  {
+                     struct hostent *HostAddr;
+
+                     HostAddr = gethostbyname (word[6]);
+                     if (HostAddr)
+                     {
+                        prefs.my_ip = ((struct in_addr *) HostAddr->h_addr)->s_addr;
+                        EMIT_SIGNAL (XP_TE_FOUNDIP, sess,
+                                     inet_ntoa (*((struct in_addr *) HostAddr->h_addr)), NULL, NULL, NULL, 0);
+                     }
                   }
                }
                break;
             case 315:          /* END OF WHO */
-               if (!serv->skip_next_who)
-                  goto def;
-               serv->skip_next_who = FALSE;
+               if (serv->skip_next_who)
+               {
+                  serv->skip_next_who = FALSE;
+               } else
+               {
+                  struct session *who_sess;
+                  who_sess = find_session_from_channel (word[4], serv);
+                  if (who_sess)
+                  {
+                     if (who_sess->doing_who)
+                        who_sess->doing_who = FALSE;
+                     else
+                        goto def;
+                  } else
+                  {
+                     if (serv->doing_who)
+                        serv->doing_who = FALSE;
+                     else
+                        goto def;
+                  }
+               }
                break;
-            case 353:
-               names_list (serv, outbuf, text);
+            case 353: /* NAMES */
+		   {
+			char *names, *chan;
+
+			chan = word[5];
+			names = word_eol[6];
+			if (*names == ':') names++;
+         		names_list (serv, outbuf, chan, names);
+		   }
                break;
             case 366:
                end_of_names (serv, outbuf, find_word (pdibuf, 4), text);
                break;
             case 376:
             case 422:          /* end of motd */
+		   serv->end_of_motd = TRUE;
                if (prefs.wallops)
                {
                   sprintf (outbuf, "MODE %s +w\r\n", serv->nick);
@@ -1466,7 +1517,7 @@ process_line (struct session *sess, struct server *serv, char *buf)
                   sprintf (outbuf, "MODE %s +i\r\n", serv->nick);
                   tcp_send (serv, outbuf);
                }
-               if (prefs.get_ip_from_server)
+               if (prefs.ip_from_server)
                {
                   serv->skip_next_who = TRUE;
                   sprintf (outbuf, "WHO %s\r\n", serv->nick);
@@ -1475,6 +1526,7 @@ process_line (struct session *sess, struct server *serv, char *buf)
                check_willjoin_channels (serv, outbuf);
                goto def;
             case 433:
+            case 437:
                next_nick (sess, outbuf, word[4]);
                break;
             case 471:
@@ -1504,7 +1556,7 @@ process_line (struct session *sess, struct server *serv, char *buf)
                      return 0;
                   }
                }
-               if (*text == '#' || *text == '&')
+               if (*text == '#' || *text == '&' || *text == '+')
                {
                   char *chan = find_word (pdibuf, 3);
                   if (!strncasecmp (serv->nick, chan, strlen (serv->nick)))
@@ -1526,7 +1578,7 @@ process_line (struct session *sess, struct server *serv, char *buf)
       } else
       {
          char t;
-         char nick[64], ip[80];
+         char nick[64], ip[128];
          char *po2, *po = strchr (buf, '!');
 
          if (po)
@@ -1560,6 +1612,8 @@ process_line (struct session *sess, struct server *serv, char *buf)
 
                if (!strcmp ("INVITE", cmd))
                {
+                  if (ignore_check (pdibuf, 0, 0, 0, 0, 1))
+                     return 0;
                   EMIT_SIGNAL (XP_TE_INVITED, sess, word[4] + 1, nick, NULL,
                              NULL, 0);
                   return 0;
@@ -1616,6 +1670,8 @@ process_line (struct session *sess, struct server *serv, char *buf)
                      char *msg = find_word_to_end (buf, 4) + 1;
                      if (*msg)
                      {
+                        if (ignore_check (pdibuf, 0, 1, 0, 0, 0))
+                           return 0;
                         notice (serv, outbuf, to, nick, msg, ip);
                         return 0;
                      }
@@ -1639,14 +1695,24 @@ process_line (struct session *sess, struct server *serv, char *buf)
                   if (*to)
                   {
                      char *msg = find_word_to_end (buf, 4) + 1;
-                     if (msg[0] == 1)  /* CTCP */
-                        handle_ctcp (sess, outbuf, to, nick, msg + 1, word, word_eol);
-                     else
+                     if (msg[0] == 1 && msg[strlen (msg) - 1] == 1)  /* ctcp */
                      {
-                        if (to[0] == '#' || to[0] == '&')
+                        if (ignore_check (pdibuf, 0, 0, 0, 1, 0))
+                           return 0;
+                        handle_ctcp (sess, outbuf, to, nick, msg + 1, word, word_eol);
+                     } else
+                     {
+                        if (to[0] == '#' || to[0] == '&' || to[0] == '+')
+                        {
+                           if (ignore_check (pdibuf, 0, 0, 1, 0, 0))
+                              return 0;
                            channel_msg (serv, outbuf, to, nick, msg, FALSE);
-                        else
+                        } else
+                        {
+                           if (ignore_check (pdibuf, 1, 0, 0, 0, 0))
+                              return 0;
                            private_msg (serv, outbuf, nick, ip, msg);
+                        }
                      }
                      return 0;
                   }
@@ -1689,23 +1755,23 @@ process_line (struct session *sess, struct server *serv, char *buf)
                              NULL, 0);
                   return 0;
                }
-
-		   if (!strcmp ("WALLOPS", cmd))
-		   {
-			char *msg = word_eol[3];
-			if(*msg == ':') msg++;
-			EMIT_SIGNAL (XP_TE_WALLOPS, sess, nick, msg, NULL,
+               if (!strcmp ("WALLOPS", cmd))
+               {
+                  char *msg = word_eol[3];
+                  if (*msg == ':')
+                     msg++;
+                  EMIT_SIGNAL (XP_TE_WALLOPS, sess, nick, msg, NULL,
                              NULL, 0);
-			return 0;
-		   }
-
+                  return 0;
+               }
                sprintf (outbuf, "(%s/%s) %s\n", nick, ip, find_word_to_end (buf, 2));
                PrintText (sess, outbuf);
                return 0;
             }
          }
-         sprintf (outbuf, STARTON " %s\n", buf);
-         PrintText (sess, outbuf);
+         /*sprintf (outbuf, STARTON " %s\n", buf);
+         PrintText (sess, outbuf);*/
+	   EMIT_SIGNAL (XP_TE_SERVERGENMESSAGE, sess, buf, NULL, NULL, NULL, 0);
       }
    }
    return 0;
