@@ -23,6 +23,7 @@
 #include "userlist.h"
 #include <gdk/gdkkeysyms.h>
 #include <fcntl.h>
+#include <arpa/inet.h>
 #ifdef USE_IMLIB
 #include <gdk_imlib.h>
 #endif
@@ -54,13 +55,17 @@ extern GdkFont *font_normal;
 extern GdkFont *font_bold;
 extern gint xchat_is_quitting;
 
+extern void palette_alloc (GtkWidget * widget);
+extern void menu_urlmenu (struct session *sess, GdkEventButton * event, char *url);
+extern void menu_hostmenu (struct session *sess, GdkEventButton * event, char *url);
+extern void menu_chanmenu (struct session *sess, GdkEventButton * event, char *url);
 extern void userlist_dnd_drop (GtkWidget *, GdkDragContext *, gint, gint, GtkSelectionData *, guint, guint32, struct session *);
 extern int userlist_dnd_motion (GtkWidget *, GdkDragContext *, gint, gint, guint);
 extern int userlist_dnd_leave (GtkWidget *, GdkDragContext *, guint time);
 extern int tcp_send_len (struct server *serv, char *buf, int len);
 extern int tcp_send (struct server *serv, char *buf);
 extern void end_logging (int fd);
-extern void menu_popup (struct session *sess, GdkEventButton * event, char *nick);
+extern void menu_popup (struct session *sess, GdkEventButton * event, char *nick, char *);
 extern void clear_user_list (struct session *sess);
 extern void xchat_cleanup (void);
 extern void kill_session_callback (GtkWidget * win, struct session *sess);
@@ -77,7 +82,10 @@ extern void popup_cmd (GtkWidget * igad, char *cmd);
 extern void path_part (char *file, char *path);
 extern struct session *new_dialog (struct session *sess);
 extern int close_dialog (GtkWidget * wid, struct session *sess);
+extern struct user *find_name (struct session *sess, char *name);
 extern void create_panel_widget ();
+extern void popup_eval_command (struct session *sess, char *cmd, char *nick, char *allnick);
+int key_handle_key_press (GtkWidget *, GdkEventKey *, gpointer);
 
 void userlist_button (GtkWidget * box, char *label, char *cmd,
                       struct session *sess, int a, int b, int c, int d);
@@ -87,15 +95,15 @@ GdkColor colors[] =
    {0, 0xffff, 0xffff, 0xffff}, /* 0  white */
    {0, 0, 0, 0},                /* 1  black */
    {0, 0, 0, 0xcccc},           /* 2  blue */
-   {0, 0, 0xbbbb, 0},           /* 3  green */
-   {0, 0xbbbb, 0, 0},           /* 4  red */
-   {0, 0xaaaa, 0xaaaa, 0},      /* 5  yellow/brown */
+   {0, 0, 0xcccc, 0},           /* 3  green */
+   {0, 0xcccc, 0, 0},           /* 4  red */
+   {0, 0xbbbb, 0xbbbb, 0},      /* 5  yellow/brown */
    {0, 0xbbbb, 0, 0xbbbb},      /* 6  purple */
-   {0, 0xcccc, 0xcccc, 0},      /* 7  orange */
-   {0, 0xeeee, 0xeeee, 0},      /* 8  yellow */
+   {0, 0xffff, 0xaaaa, 0},      /* 7  orange */
+   {0, 0xffff, 0xffff, 0},      /* 8  yellow */
    {0, 0, 0xffff, 0},           /* 9  green */
    {0, 0, 0xbbbb, 0xbbbb},      /* 10 aqua */
-   {0, 0, 0xeeee, 0xeeee},      /* 11 light aqua */
+   {0, 0, 0xffff, 0xffff},      /* 11 light aqua */
    {0, 0, 0, 0xffff},           /* 12 blue */
    {0, 0xffff, 0, 0xffff},      /* 13 pink */
    {0, 0x7777, 0x7777, 0x7777}, /* 14 grey */
@@ -169,7 +177,7 @@ gui_set_title (struct session *sess)
       strcpy (tbuf, "X-Chat (" VERSION ")");
    else
    {
-      if (sess->channel[0] == 0)
+      if (sess->channel[0] == 0 || sess->is_server)
          snprintf (tbuf, sizeof tbuf, "X-Chat (" VERSION "): %s", sess->server->servername);
       else
          snprintf (tbuf, sizeof tbuf, "X-Chat (" VERSION "): %s / %s", sess->server->servername, sess->channel);
@@ -223,7 +231,10 @@ clear_channel (struct session *sess)
       gtk_widget_destroy (sess->op_xpm);
    sess->op_xpm = 0;
    if (sess->logfd != -1)
+   {
       end_logging (sess->logfd);
+	sess->logfd = -1;
+   }
 }
 
 void
@@ -273,17 +284,28 @@ find_selected_nick (struct session *sess)
 void
 ul_button_rel (GtkWidget * widget, GdkEventButton * even, struct session *sess)
 {
+   struct user *user;
    char *nick;
    int row, col;
-   char buf[64];
+   char buf[67];
 
    if (!even)
       return;
    if (even->button == 3)
    {
+      if (gtk_clist_get_selection_info (GTK_CLIST (widget), even->x, even->y, &row, &col) < 0)
+         return;
+	gtk_clist_unselect_all (GTK_CLIST (widget));
+	gtk_clist_select_row (GTK_CLIST (widget), row, 0);
       nick = find_selected_nick (sess);
       if (nick)
-         menu_popup (sess, even, nick);
+      {
+         user = find_name (sess, nick);
+         if (user)
+            menu_popup (sess, even, nick, user->hostname);
+         else
+            menu_popup (sess, even, nick, 0);
+      }
       return;
    }
    if (even->button == 2)
@@ -316,14 +338,22 @@ focus_in (GtkWindow * win, GtkWidget * wid, struct session *sess)
    {
       if (current_tab)
       {
-         gtk_widget_grab_focus (current_tab->inputgad);
+         if (!current_tab->is_shell)
+            gtk_widget_grab_focus (current_tab->inputgad);
+         else
+            gtk_widget_grab_focus (current_tab->textgad);
          menu_sess = current_tab;
-         current_tab->server->front_session = current_tab;
+	   if (!prefs.use_server_tab)
+      	current_tab->server->front_session = current_tab;
       }
    } else
    {
-      sess->server->front_session = sess;
-      gtk_widget_grab_focus (sess->inputgad);
+	if (!prefs.use_server_tab)
+   	   sess->server->front_session = sess;
+      if (!sess->is_shell)
+         gtk_widget_grab_focus (sess->inputgad);
+      else
+         gtk_widget_grab_focus (sess->textgad);
       menu_sess = sess;
 #ifdef USE_PANEL
       if (sess->panel_button)
@@ -508,12 +538,14 @@ userlist_button (GtkWidget * box, char *label, char *cmd,
                  struct session *sess, int a, int b, int c, int d)
 {
    GtkWidget *wid = gtk_button_new_with_label (label);
+   gtk_object_set_user_data (GTK_OBJECT (wid), sess);
    gtk_signal_connect (GTK_OBJECT (wid), "clicked",
     GTK_SIGNAL_FUNC (popup_cmd), cmd);
    gtk_table_attach_defaults (GTK_TABLE (box), wid, a, b, c, d);
    show_and_unfocus (wid);
 }
 
+
 #ifdef USE_GNOME
 
 void
@@ -552,7 +584,7 @@ zvt_set_style (GtkWidget * zvt, char *bg_pic, int trans, int tint,
 #endif
 
 GtkStyle *
-my_widget_get_style (char *bg_pic, int bg_color)
+my_widget_get_style (char *bg_pic)
 {
    char buf[256];
    GtkStyle *style;
@@ -563,8 +595,8 @@ my_widget_get_style (char *bg_pic, int bg_color)
 
    style = gtk_style_new ();
 
-   style->base[GTK_STATE_NORMAL] = colors[bg_color];
-   style->bg[GTK_STATE_NORMAL] = colors[bg_color];
+   style->base[GTK_STATE_NORMAL] = colors[1];
+   style->bg[GTK_STATE_NORMAL] = colors[1];
 
    if (bg_pic[0])
    {
@@ -615,20 +647,10 @@ textgad_get_focus_cb (GtkWidget * wid, GdkEventKey * event, struct session *sess
 }
 
 #ifdef USE_GNOME
-/* borrowed from gnome-terminal */
 
-gushort rxvt_red[] =
-{0x0000, 0xffff, 0x0000, 0xffff, 0x0000, 0xffff, 0x0000, 0xffff,
- 0x7777, 0xffff, 0x0000, 0xffff, 0x0000, 0xffff, 0x0000, 0xffff,
- 0xffff, 0x0};
-gushort rxvt_grn[] =
-{0x0000, 0x0000, 0xffff, 0xffff, 0x0000, 0x0000, 0xffff, 0xffff,
- 0x7777, 0x0000, 0xffff, 0xffff, 0x0000, 0x0000, 0xffff, 0xffff,
- 0xffff, 0x0};
-gushort rxvt_blu[] =
-{0x0000, 0x0000, 0x0000, 0x0000, 0xffff, 0xffff, 0xffff, 0xffff,
- 0x7777, 0x0000, 0x0000, 0x0000, 0xffff, 0xffff, 0xffff, 0xffff,
- 0xffff, 0x0};
+gushort zvt_red[18];
+gushort zvt_grn[18];
+gushort zvt_blu[18];            /* these are filled in palette.c palette_prepare() */
 
 void
 term_change_pos (GtkWidget * widget)
@@ -638,6 +660,281 @@ term_change_pos (GtkWidget * widget)
 
 #endif
 
+static int
+is_delimiter (char c)
+{
+   switch (c)
+   {
+   case ' ':
+   case '|':
+   case '[':
+   case ']':
+   case '(':
+   case ')':
+   case '<':
+   case '>':
+   case '\0':
+   case '\n':
+   case '"':
+      return 1;
+   }
+   return 0;
+}
+
+/* mouse click inside text area */
+
+static void
+maingui_word_clicked (struct session *sess, GdkEventButton * even, char *word)
+{
+   struct user *user;
+   int url = 0;
+
+   if (even && word[0] && (word[0] != ' ' && word[1] != 0))
+   {
+      if (even->button == 3)
+      {
+	   if(word[0] == '#')
+	   {
+		menu_chanmenu (sess, even, word);
+		return;
+	   }
+         user = find_name (sess, word);
+         if (user)              /* see if this word is in the userlist */
+         {
+            menu_popup (sess, even, word, user->hostname);
+            return;
+         } else
+         {
+            if (!strncasecmp (word, "ftp.", 4))
+               url = 1;
+            if (!strncasecmp (word, "ftp:", 4))
+               url = 1;
+            if (!strncasecmp (word, "www.", 4))
+               url = 1;
+            if (!strncasecmp (word, "http:", 5))
+               url = 1;
+            if (!url)
+            {
+               char *at = strchr (word, '@');  /* check for email addy */
+               char *dot = strrchr (word, '.');
+               if (at && dot)
+               {
+                  if ((unsigned long) at < (unsigned long) dot)
+                  {
+                     char *newword = malloc (strlen (word) + 10);
+                     if (*word == '~')
+                        word++;
+                     sprintf (newword, "mailto:%s", word);
+                     menu_urlmenu (sess, even, newword);
+                     free (newword);
+                     return;
+                  }
+               }
+            }
+            if (url)
+            {
+               menu_urlmenu (sess, even, word);
+               return;
+            } else
+            {
+               char *dot = strrchr (word, '.');
+               if (dot)
+               {
+                  dot++;
+                  if (!strcasecmp (dot, "org"))
+                     url = 1;
+                  else if (!strcasecmp (dot, "com"))
+                     url = 1;
+                  else if (!strcasecmp (dot, "net"))
+                     url = 1;
+                  else if (!strcasecmp (dot, "edu"))
+                     url = 1;
+			else if (word[strlen(word)-3] == '.')
+			   url = 1;
+                  if (url)
+                  {
+                     menu_hostmenu (sess, even, word);
+                     return;
+                  }
+               }
+               {
+                  int i, dots = 0;  /* check if it's an IP number */
+
+                  for (i = 0; i < strlen (word); i++)
+                  {
+                     if (word[i] == '.')
+                        dots++;
+                  }
+                  if (dots == 3)
+                  {
+                     if (inet_addr (word) != -1)
+                     {
+                        menu_hostmenu (sess, even, word);
+                        return;
+                     }
+                  }
+               }
+            }
+         }
+      } else if (even->button == 4)
+      {
+         GtkAdjustment *adj;
+         gfloat new_value;
+         adj = gtk_range_get_adjustment (GTK_RANGE (sess->vscrollbar));
+         new_value = adj->value - adj->page_increment;
+         if (new_value < adj->lower)
+            new_value = adj->lower;
+         gtk_adjustment_set_value (adj, new_value);
+      } else if (even->button == 5)
+      {
+         GtkAdjustment *adj;
+         gfloat new_value;
+         adj = gtk_range_get_adjustment (GTK_RANGE (sess->vscrollbar));
+         new_value = adj->value + adj->page_increment;
+         if (new_value > (adj->upper - adj->page_size))
+            new_value = adj->upper - adj->page_size;
+         gtk_adjustment_set_value (adj, new_value);
+      }
+   }
+   if (even && even->button == 3 && sess->is_dialog)
+   {
+      menu_popup (sess, even, sess->channel, gtk_entry_get_text (GTK_ENTRY (sess->topicgad)));
+   }
+}
+
+#ifdef USE_GNOME
+
+gint
+maingui_textgad_clicked_zvt (GtkWidget * wid, GdkEventButton * even, struct session *sess)
+{
+   char *text, *text2, *wordstart, *word;
+   int font_x, font_y, len, len2, pos, end, cur;
+   GtkAdjustment *adj;
+
+   if (even->button == 3)
+      gtk_signal_emit_stop_by_name (GTK_OBJECT (wid), "button_press_event");
+
+   adj = ((ZvtTerm *) sess->textgad)->adjustment;
+   end = adj->upper - adj->lower - adj->page_size;
+   cur = adj->value + (adj->page_size / 2);
+   if (cur < end)
+   {
+      /* we're not on the last page, so forget about it. */
+      /* FIXME: maybe use the adjustment to offset font_y negatively?? */
+      return TRUE;
+   }
+   font_x = even->x / ((ZvtTerm *) sess->textgad)->charwidth;
+   font_y = even->y / ((ZvtTerm *) sess->textgad)->charheight;
+
+   text = zvt_term_get_buffer ((ZvtTerm *) sess->textgad, &len, VT_SELTYPE_CHAR, 0,
+                               font_y, ((ZvtTerm *) sess->textgad)->grid_width, font_y);
+
+   if (!text || !len)
+      return TRUE;
+
+   /* let's get the next line too, and join them, incase the url is
+      split across 2 lines. */
+   text2 = zvt_term_get_buffer ((ZvtTerm *) sess->textgad, &len2, VT_SELTYPE_CHAR, 0,
+                                font_y + 1, ((ZvtTerm *) sess->textgad)->grid_width - 1, font_y + 1);
+
+   if (text2 && len2)
+   {
+      char *newtext = g_malloc (len + len2 + 2);
+      memcpy (newtext, text, len);
+      memcpy (newtext + len, text2, len2);
+      newtext[len + len2] = 0;
+      g_free (text);
+      g_free (text2);
+      text = newtext;
+   }
+   pos = font_x;
+   while (!is_delimiter (text[pos]))
+      pos--;
+
+   pos++;
+   wordstart = &text[pos];
+
+   len = 0;
+   while (!is_delimiter (text[pos]))
+   {
+      pos++;
+      len++;
+   }
+
+   word = malloc (len + 1);
+   memcpy (word, wordstart, len);
+   word[len] = 0;
+   g_free (text);
+
+   maingui_word_clicked (sess, even, word);
+
+   free (word);
+
+   return TRUE;
+}
+
+#endif
+
+gint
+maingui_textgad_clicked_gtk (GtkWidget * wid, GdkEventButton * even, struct session *sess)
+{
+   char *text, *word, *wordstart;
+   int len, pos, from, to;
+
+   if (even->button == 3)
+      gtk_signal_emit_stop_by_name (GTK_OBJECT (wid), "button_press_event");
+
+   len = gtk_text_get_length (GTK_TEXT (sess->textgad));
+   pos = gtk_editable_get_position (GTK_EDITABLE (sess->textgad));
+
+   if (pos >= len)
+   {
+	if(sess->is_dialog)
+   	   maingui_word_clicked (sess, even, "");
+      return TRUE;
+   }
+
+   if (pos < 100)
+      from = 0;
+   else
+      from = pos - 100;
+
+   if (pos + 100 > len)
+      to = len;
+   else
+      to = pos + 100;
+
+   text = gtk_editable_get_chars (GTK_EDITABLE (sess->textgad), from, to);
+
+   if (pos > 100)
+      pos = 100;
+
+   while (!is_delimiter (text[pos]))
+      pos--;
+
+   pos++;
+   wordstart = &text[pos];
+
+   len = 0;
+   while (!is_delimiter (text[pos]))
+   {
+      pos++;
+      len++;
+   }
+
+   word = malloc (len + 1);
+   memcpy (word, wordstart, len);
+   word[len] = 0;
+   g_free (text);
+
+   maingui_word_clicked (sess, even, word);
+
+   free (word);
+
+   return TRUE;
+}
+
+
 void
 maingui_create_textlist (struct session *sess, GtkWidget * leftpane)
 {
@@ -659,14 +956,17 @@ maingui_create_textlist (struct session *sess, GtkWidget * leftpane)
        prefs.mainwindow_height - 105);
       gtk_signal_connect (GTK_OBJECT (sess->textgad), "key_press_event",
                           GTK_SIGNAL_FUNC (textgad_get_focus_cb), sess);
+      gtk_signal_connect (GTK_OBJECT (sess->textgad), "button_press_event",
+                          GTK_SIGNAL_FUNC (maingui_textgad_clicked_zvt), sess);
       zvt_term_set_scrollback ((ZvtTerm *) sess->textgad, prefs.zvt_lines);
+	zvt_term_set_blink ((ZvtTerm *) sess->textgad, FALSE);
       zvt_set_style (sess->textgad, prefs.background, prefs.transparent,
                      prefs.tint, font_normal, font_bold);
       gtk_container_add (GTK_CONTAINER (leftpane), sess->textgad);
       gtk_widget_show (sess->textgad);
 
       sess->vscrollbar = gtk_vscrollbar_new (((ZvtTerm *) sess->textgad)->adjustment);
-      gtk_box_pack_start (GTK_BOX (leftpane), sess->vscrollbar, FALSE, FALSE, 0);
+      gtk_box_pack_start (GTK_BOX (leftpane), sess->vscrollbar, FALSE, FALSE, 1);
       show_and_unfocus (sess->vscrollbar);
    } else
    {
@@ -675,17 +975,18 @@ maingui_create_textlist (struct session *sess, GtkWidget * leftpane)
          gtk_widget_realize (sess->window);
       sess->textgad = gtk_text_new (0, 0);
       gtk_widget_set_style (sess->textgad, channelwin_style);
-      gtk_widget_set_usize (sess->textgad,
-         prefs.mainwindow_width - 115,
-       prefs.mainwindow_height - 105);
+      gtk_widget_set_usize (sess->textgad, prefs.mainwindow_width - 115,
+							 prefs.mainwindow_height - 105);
       gtk_signal_connect (GTK_OBJECT (sess->textgad), "key_press_event",
                           GTK_SIGNAL_FUNC (textgad_get_focus_cb), sess);
+      gtk_signal_connect (GTK_OBJECT (sess->textgad), "button_press_event",
+                          GTK_SIGNAL_FUNC (maingui_textgad_clicked_gtk), sess);
       gtk_text_set_word_wrap (GTK_TEXT (sess->textgad), TRUE);
       gtk_container_add (GTK_CONTAINER (leftpane), sess->textgad);
       gtk_widget_show (sess->textgad);
 
       sess->vscrollbar = gtk_vscrollbar_new (GTK_TEXT (sess->textgad)->vadj);
-      gtk_box_pack_start (GTK_BOX (leftpane), sess->vscrollbar, FALSE, FALSE, 0);
+      gtk_box_pack_start (GTK_BOX (leftpane), sess->vscrollbar, FALSE, FALSE, 1);
       show_and_unfocus (sess->vscrollbar);
 #ifdef USE_GNOME
    }
@@ -697,9 +998,12 @@ gui_new_tab (struct session *sess)
 {
    current_tab = sess;
    menu_sess = sess;
-   sess->server->front_session = sess;
+   if (!prefs.use_server_tab)
+	sess->server->front_session = sess;
    gui_set_title (sess);
-   gtk_widget_grab_focus (sess->inputgad);
+   if (!sess->is_shell)
+	gtk_widget_grab_focus (sess->inputgad);
+
 #ifdef USE_PANEL
    if (sess->panel_button)
       gtk_widget_set_style (GTK_BIN (sess->panel_button)->child, normaltab_style);
@@ -733,6 +1037,13 @@ gui_new_tab_callback (GtkWidget * widget, GtkNotebookPage * nbpage, guint page)
    }
 }
 
+int
+maingui_pagetofront (int page)
+{
+     gtk_notebook_set_page (GTK_NOTEBOOK (main_book), page);
+     return 0;
+}
+
 void
 gui_mainbook_invalid (GtkWidget * w, GtkWidget * main_window)
 {
@@ -801,12 +1112,31 @@ void
 maingui_userlist_selected (GtkWidget * clist, gint row, gint column,
                 GdkEventButton * even)
 {
-   if (even && even->type == GDK_2BUTTON_PRESS)
+   char *nick;
+
+   if (even)
    {
-      if (prefs.doubleclickuser[0])
-      {
-         popup_cmd (0, prefs.doubleclickuser);
-      }
+	if (even->type == GDK_2BUTTON_PRESS)
+	{
+   	   if (prefs.doubleclickuser[0])
+   	   {
+
+      	if (gtk_clist_get_cell_type (GTK_CLIST (clist), row, column) == GTK_CELL_PIXTEXT)
+      	{
+         	   void *unused;
+         	   gtk_clist_get_pixtext (GTK_CLIST (clist), row, column, &nick, (guint8 *) & unused, (GdkPixmap **) & unused, (GdkBitmap **) & unused);
+      	} else
+         	   gtk_clist_get_text (GTK_CLIST (clist), row, column, &nick);
+      	popup_eval_command (menu_sess, prefs.doubleclickuser, nick, nick);
+   	   }
+	} else
+	{
+	   if(!(even->state & GDK_SHIFT_MASK))
+	   {
+		gtk_clist_unselect_all (GTK_CLIST (clist));
+		gtk_clist_select_row (GTK_CLIST (clist), row, column);
+	   }
+	}
    }
 }
 
@@ -968,7 +1298,7 @@ relink_window (GtkWidget * w, struct session *sess)
 #ifdef USE_GNOME
       if (sess->zvt)
       {
-         zvt_term_set_color_scheme ((ZvtTerm *) sess->textgad, rxvt_red, rxvt_grn, rxvt_blu);
+         zvt_term_set_color_scheme ((ZvtTerm *) sess->textgad, zvt_red, zvt_grn, zvt_blu);
          gtk_widget_hide (sess->textgad);
          gtk_widget_show (sess->textgad);
       }
@@ -1019,7 +1349,7 @@ relink_window (GtkWidget * w, struct session *sess)
          gtk_widget_show_all (main_window);
 #ifdef USE_GNOME
       if (sess->zvt)
-         zvt_term_set_color_scheme ((ZvtTerm *) sess->textgad, rxvt_red, rxvt_grn, rxvt_blu);
+         zvt_term_set_color_scheme ((ZvtTerm *) sess->textgad, zvt_red, zvt_grn, zvt_blu);
 #endif
       return 0;
    }
@@ -1049,13 +1379,10 @@ gui_panel_remove_clicked (GtkWidget * button, struct session *sess)
    sess->panel_button = 0;
    if (main_window)             /* this fixes a little refresh glitch */
    {
-      int num;
+      int page;
 
-      while (gtk_events_pending ())
-         gtk_main_iteration ();
-
-      num = gtk_notebook_page_num (GTK_NOTEBOOK (main_book), sess->window);
-      gtk_notebook_set_page (GTK_NOTEBOOK (main_book), num);
+      page = gtk_notebook_page_num (GTK_NOTEBOOK (main_book), sess->window);
+	gtk_idle_add ((GtkFunction)maingui_pagetofront, (gpointer)page);
    }
    if (panel_popup)
       gui_panel_destroy_popup (NULL, NULL);
@@ -1112,7 +1439,7 @@ gui_panel_view_clicked (GtkWidget * button, struct session *sess)
    gtk_widget_reparent (sess->textgad, view_popup);
 #ifdef USE_GNOME
    if (sess->zvt)
-      zvt_term_set_color_scheme ((ZvtTerm *) sess->textgad, rxvt_red, rxvt_grn, rxvt_blu);
+      zvt_term_set_color_scheme ((ZvtTerm *) sess->textgad, zvt_red, zvt_grn, zvt_blu);
 #endif
    gtk_signal_connect (GTK_OBJECT (view_popup), "leave_notify_event", gui_panel_destroy_view_popup, sess);
 }
@@ -1195,13 +1522,10 @@ maingui_unpanelize (GtkWidget * button, struct session *sess)
       sess->panel_button = 0;
    if (main_window)             /* this fixes a little refresh glitch */
    {
-      int num;
+      int page;
 
-      while (gtk_events_pending ())
-         gtk_main_iteration ();
-
-      num = gtk_notebook_page_num (GTK_NOTEBOOK (main_book), sess->window);
-      gtk_notebook_set_page (GTK_NOTEBOOK (main_book), num);
+      page = gtk_notebook_page_num (GTK_NOTEBOOK (main_book), sess->window);
+	gtk_idle_add ((GtkFunction)maingui_pagetofront, (gpointer)page);
    }
 }
 
@@ -1239,13 +1563,35 @@ maingui_panelize (GtkWidget * button, struct session *sess)
 
 #endif
 
+/* 'X' button pressed */
+
+void
+X_session (GtkWidget * button, struct session *sess)
+{
+   gtk_widget_destroy (sess->window);
+}
+
+void
+maingui_init_styles (GtkStyle *style)
+{
+   normaltab_style = gtk_style_new ();
+   normaltab_style->font = style->font;
+
+   redtab_style = gtk_style_new ();
+   redtab_style->font = style->font;
+   memcpy (redtab_style->fg, &colors[4], sizeof (GdkColor));
+
+   bluetab_style = gtk_style_new ();
+   bluetab_style->font = style->font;
+   memcpy (bluetab_style->fg, &colors[12], sizeof (GdkColor));
+}
+
 void
 create_window (struct session *sess)
 {
    GtkWidget *leftpane, *rightpane;
-   GtkWidget *vbox, *tbox, *bbox, *nlbox;
-   GtkWidget *wid;
-   int i, justopened = FALSE;
+   GtkWidget *vbox, *tbox, *bbox, *nlbox, *wid;
+   int justopened = FALSE;
 
    if (sess->is_dialog)
    {
@@ -1287,17 +1633,8 @@ create_window (struct session *sess)
       gtk_window_set_policy ((GtkWindow *) sess->window, TRUE, TRUE, FALSE);
    }
 
-   if (colors[0].pixel == 0)    /* only do this the first time through */
-   {
-      for (i = 0; i < 16; i++)
-      {
-         colors[i].pixel =
-            (gulong) ((colors[i].red & 0xff00) * 256 +
-          (colors[i].green & 0xff00) +
-                      (colors[i].blue & 0xff00) / 256);
-         gdk_color_alloc (gtk_widget_get_colormap (sess->window), &colors[i]);
-      }
-   }
+   palette_alloc (sess->window);
+
    vbox = gtk_vbox_new (FALSE, 0);
    sess->vbox = vbox;
    gtk_container_set_border_width ((GtkContainer *) vbox, 2);
@@ -1331,36 +1668,37 @@ create_window (struct session *sess)
    gtk_widget_show (tbox);
 
 #ifdef USE_GNOME
-   wid = gtkutil_stock_button (sess->window, GNOME_STOCK_BUTTON_CANCEL,
-                               0, gtkutil_destroy, sess->window, 0);
+   wid = gtkutil_button (sess->window, GNOME_STOCK_BUTTON_CANCEL,
+               0, X_session, sess, 0);
    gtk_box_pack_start (GTK_BOX (tbox), wid, 0, 0, 0);
    gtk_widget_show (wid);
+   add_tip (wid, "Close Channel");
+
+   wid = gtkutil_button (sess->window, GNOME_STOCK_BUTTON_UP,
+           0, relink_window, sess, 0);
+   gtk_box_pack_start (GTK_BOX (tbox), wid, 0, 0, 0);
+   gtk_widget_show (wid);
+   add_tip (wid, "Link/DeLink this tab");
 #else
    wid = gtk_button_new_with_label ("X");
    gtk_box_pack_start (GTK_BOX (tbox), wid, 0, 0, 0);
    gtk_signal_connect (GTK_OBJECT (wid), "clicked",
-                       GTK_SIGNAL_FUNC (gtkutil_destroy), sess->window);
+                       GTK_SIGNAL_FUNC (X_session), sess);
    gtk_widget_show (wid);
-#endif
    add_tip (wid, "Close Channel");
 
-#ifdef USE_GNOME
-   wid = gtkutil_stock_button (sess->window, GNOME_STOCK_BUTTON_UP,
-           0, relink_window, sess, 0);
-   gtk_box_pack_start (GTK_BOX (tbox), wid, 0, 0, 0);
-#else
    wid = gtk_button_new_with_label ("^");
    gtk_box_pack_start (GTK_BOX (tbox), wid, 0, 0, 0);
    gtk_signal_connect (GTK_OBJECT (wid), "clicked",
                        GTK_SIGNAL_FUNC (relink_window), sess);
-#endif
    gtk_widget_show (wid);
    add_tip (wid, "Link/DeLink this tab");
+#endif
 
 #ifdef USE_PANEL
    if (!nopanel)
    {
-      wid = gtkutil_stock_button (sess->window, GNOME_STOCK_BUTTON_DOWN,
+      wid = gtkutil_button (sess->window, GNOME_STOCK_BUTTON_DOWN,
         0, maingui_panelize, sess, 0);
       gtk_box_pack_start (GTK_BOX (tbox), wid, 0, 0, 0);
       add_tip (wid, "Panelize");
@@ -1385,10 +1723,10 @@ create_window (struct session *sess)
    {
 #ifdef USE_GNOME
       if (sess->is_tab)
-         wid = gtkutil_stock_button (main_window, GNOME_STOCK_PIXMAP_FORWARD, 0,
+         wid = gtkutil_button (main_window, GNOME_STOCK_PIXMAP_FORWARD, 0,
               userlist_hide, sess, 0);
       else
-         wid = gtkutil_stock_button (sess->window, GNOME_STOCK_PIXMAP_FORWARD, 0,
+         wid = gtkutil_button (sess->window, GNOME_STOCK_PIXMAP_FORWARD, 0,
               userlist_hide, sess, 0);
       gtk_box_pack_end (GTK_BOX (tbox), wid, 0, 0, 0);
 #else
@@ -1478,8 +1816,8 @@ create_window (struct session *sess)
    sess->leftpane = leftpane;
 
    sess->namelistgad = gtkutil_clist_new (1, 0, nlbox, GTK_POLICY_AUTOMATIC,
-      maingui_userlist_selected, sess,
-          0, 0, GTK_SELECTION_BROWSE);
+   maingui_userlist_selected, sess, 0,
+           0, GTK_SELECTION_MULTIPLE);
 #ifdef USE_GNOME
    /* set up drops */
    gtk_drag_dest_set (sess->namelistgad, GTK_DEST_DEFAULT_ALL,
@@ -1492,7 +1830,8 @@ create_window (struct session *sess)
                        GTK_SIGNAL_FUNC (userlist_dnd_drop), sess);
 #endif
 
-   gtk_clist_set_selection_mode (GTK_CLIST (sess->namelistgad), GTK_SELECTION_SINGLE);
+   gtk_clist_set_selection_mode (GTK_CLIST (sess->namelistgad),
+              GTK_SELECTION_MULTIPLE);
    gtk_clist_set_column_width (GTK_CLIST (sess->namelistgad), 0, 10);
    gtk_widget_set_usize (sess->namelistgad->parent, 115, 0);
    gtk_signal_connect (GTK_OBJECT (sess->namelistgad), "button_press_event",
@@ -1521,18 +1860,29 @@ create_window (struct session *sess)
    gtk_signal_connect (GTK_OBJECT (sess->inputgad), "activate",
                        GTK_SIGNAL_FUNC (handle_inputgad), sess);
    gtk_signal_connect (GTK_OBJECT (sess->inputgad), "key_press_event",
-                       GTK_SIGNAL_FUNC (history_keypress), &sess->history);
+                       GTK_SIGNAL_FUNC (key_handle_key_press), sess);
    gtk_widget_show (sess->inputgad);
-   gtk_widget_grab_focus (sess->inputgad);
-   /*add_tip(sess->inputgad, "Type in here"); */
+   if (prefs.newtabstofront || justopened)
+      gtk_widget_grab_focus (sess->inputgad);
+
+   sess->confbutton = wid = gtk_toggle_button_new_with_label ("C");
+   gtk_box_pack_end (GTK_BOX (bbox), wid, 0, 0, 1);
+   gtk_widget_show (wid);
+   add_tip (wid, "Conference mode (no join/part msgs)");
 
    gtk_widget_show (sess->window);
 
+   if (!normaltab_style && prefs.tabchannels)
+   {
+	if (prefs.mainwindow_left || prefs.mainwindow_top)
+	   gdk_window_move (main_window->window,
+                          prefs.mainwindow_left,
+                          prefs.mainwindow_top);
+   }
+
    if (prefs.tabchannels)
    {
-      GtkStyle *style;
-
-      sess->changad = gtk_label_new ("<none>");
+	sess->changad = gtk_label_new ("<none>");
       gtk_widget_show (sess->changad);
 
       gtk_notebook_append_page (GTK_NOTEBOOK (main_book), sess->window, sess->changad);
@@ -1540,51 +1890,42 @@ create_window (struct session *sess)
       if (justopened)
          gtk_widget_show (main_window);
 
-      style = gtk_widget_get_style (sess->changad);
+	if (!normaltab_style)
+	   maingui_init_styles (gtk_widget_get_style(sess->changad));
 
-      if (!normaltab_style)
+#ifdef USE_GNOME
+      if (prefs.zvt)
       {
-         normaltab_style = gtk_style_new ();
-         normaltab_style->font = style->font;
-
-         redtab_style = gtk_style_new ();
-         redtab_style->font = style->font;
-         memcpy (redtab_style->fg, &colors[4], sizeof (GdkColor));
-
-         bluetab_style = gtk_style_new ();
-         bluetab_style->font = style->font;
-         memcpy (bluetab_style->fg, &colors[12], sizeof (GdkColor));
-
-         if (prefs.mainwindow_left || prefs.mainwindow_top)
-            gdk_window_move (main_window->window,
-                prefs.mainwindow_left,
-                prefs.mainwindow_top);
+         zvt_term_set_color_scheme ((ZvtTerm *) sess->textgad, zvt_red, zvt_grn, zvt_blu);
+         gtk_widget_hide (sess->textgad);
+         gtk_widget_show (sess->textgad);
       }
+#endif
+
       if (prefs.newtabstofront)
       {
-         int num;
+         int page;
 
-         while (gtk_events_pending ())
-            gtk_main_iteration ();
-
-         num = gtk_notebook_page_num (GTK_NOTEBOOK (main_book), sess->window);
-
-         gtk_notebook_set_page (GTK_NOTEBOOK (main_book), num);
+         page = gtk_notebook_page_num (GTK_NOTEBOOK (main_book), sess->window);
+	   gtk_idle_add ((GtkFunction)maingui_pagetofront, (gpointer)page);
 
          sess->new_data = TRUE;
          gui_new_tab (sess);
       }
    } else
    {
+	if (!normaltab_style)
+   	   maingui_init_styles (gtk_widget_get_style(sess->changad));
+
       if (prefs.mainwindow_left || prefs.mainwindow_top)
          gdk_window_move (sess->window->window,
                 prefs.mainwindow_left,
                 prefs.mainwindow_top);
    }
 #ifdef USE_GNOME
-   if (prefs.zvt)
+   if (prefs.zvt && !prefs.tabchannels)
    {
-      zvt_term_set_color_scheme ((ZvtTerm *) sess->textgad, rxvt_red, rxvt_grn, rxvt_blu);
+      zvt_term_set_color_scheme ((ZvtTerm *) sess->textgad, zvt_red, zvt_grn, zvt_blu);
       gtk_widget_hide (sess->textgad);
       gtk_widget_show (sess->textgad);
    }
